@@ -42,15 +42,6 @@
     POCL_WARN_INCOMPLETE ();                                                  \
   POCL_RETURN_GETINFO (__TYPE__, __VALUE__);
 
-#define STRINGIFY_(x) #x
-#define STRINGIFY(x) STRINGIFY_ (x)
-#define HOST_DEVICE_CL_VERSION_MAJOR_STR                                      \
-  STRINGIFY (HOST_DEVICE_CL_VERSION_MAJOR)
-#define HOST_DEVICE_CL_VERSION_MINOR_STR                                      \
-  STRINGIFY (HOST_DEVICE_CL_VERSION_MINOR)
-#define HOST_CL_VERSION                                                       \
-  "OpenCL C " HOST_DEVICE_CL_VERSION_MAJOR_STR                                \
-  "." HOST_DEVICE_CL_VERSION_MINOR_STR " pocl"
 
 CL_API_ENTRY cl_int CL_API_CALL
 POname(clGetDeviceInfo)(cl_device_id   device,
@@ -74,22 +65,6 @@ POname(clGetDeviceInfo)(cl_device_id   device,
   case CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS          :
     POCL_RETURN_GETINFO(cl_uint, device->max_work_item_dimensions);
   case CL_DEVICE_MAX_WORK_GROUP_SIZE               : 
-    /* There is no "preferred WG size" device query, so we probably should
-       return something more sensible than the CL_INT_MAX that seems
-       to be the default in the pthread device. It should be computed from 
-       the machine's vector width or issue width.
-
-       Some OpenCL programs (e.g. the Dijkstra book sample) seem to scale 
-       the work groups using this. 
-
-       There's a kernel query CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE
-       that can yield better heuristics for the good WG size and forms
-       a basis for a higher level performance portability layer.
-       
-       Basically the size is now limited by the absence of work item
-       loops. A huge unrolling factor explodes the instruction memory size (and
-       compilation time) with usually no benefits.
-    */
     {
       size_t max_wg_size = device->max_work_group_size;
       POCL_RETURN_GETINFO(size_t, max_wg_size);
@@ -150,6 +125,13 @@ POname(clGetDeviceInfo)(cl_device_id   device,
   case CL_DEVICE_MAX_SAMPLERS:
     POCL_RETURN_DEVICE_INFO_WITH_IMG_CHECK (cl_uint, device->max_samplers);
 
+  case CL_DEVICE_IMAGE_PITCH_ALIGNMENT:
+    /* Creating a 2D image from a buffer is not supported */
+    POCL_RETURN_GETINFO(cl_uint, 0);
+  case CL_DEVICE_IMAGE_BASE_ADDRESS_ALIGNMENT:
+    /* Creating a 2D image from a buffer is not supported */
+    POCL_RETURN_GETINFO(cl_uint, 0);
+
   case CL_DEVICE_MAX_PARAMETER_SIZE:
     POCL_RETURN_DEVICE_INFO_WITH_IMPL_CHECK(size_t, device->max_parameter_size);
   case CL_DEVICE_MEM_BASE_ADDR_ALIGN               :
@@ -240,8 +222,16 @@ POname(clGetDeviceInfo)(cl_device_id   device,
     POCL_RETURN_DEVICE_INFO_WITH_EXT_CHECK(cl_uint, device->native_vector_width_double, cl_khr_fp64);
   case CL_DEVICE_NATIVE_VECTOR_WIDTH_HALF          : 
     POCL_RETURN_DEVICE_INFO_WITH_EXT_CHECK(cl_uint, device->native_vector_width_half, cl_khr_fp16);
+  /* Returns a fixed (1.2) version for all devices. The spec says:
+   * .. highest fully backwards compatible OpenCL C version supported by
+   * the compiler. An OpenCL 3.0 device may return an OpenCL C version newer
+   * than OpenCL C 1.2 if and only if all optional OpenCL C features are
+   * supported by the device for the newer version.
+   *
+   * none of the PoCL devices have fully compatible 2.x compilers,
+   * and also in 3.0 this query is deprecated (there is a better solution). */
   case CL_DEVICE_OPENCL_C_VERSION                  :
-    POCL_RETURN_GETINFO_STR (HOST_CL_VERSION);
+    POCL_RETURN_GETINFO_STR ("OpenCL C 1.2 PoCL");
   case CL_DEVICE_BUILT_IN_KERNELS                  :
     if (device->builtin_kernel_list)
       POCL_RETURN_GETINFO_STR (device->builtin_kernel_list);
@@ -285,10 +275,6 @@ POname(clGetDeviceInfo)(cl_device_id   device,
 
   case CL_DEVICE_SVM_CAPABILITIES:
     POCL_RETURN_GETINFO(cl_device_svm_capabilities, device->svm_caps);
-  case CL_DEVICE_ATOMIC_MEMORY_CAPABILITIES:
-    POCL_RETURN_GETINFO(cl_device_atomic_capabilities, device->atomic_memory_capabilities);
-  case CL_DEVICE_ATOMIC_FENCE_CAPABILITIES:
-    POCL_RETURN_GETINFO(cl_device_atomic_capabilities, device->atomic_fence_capabilities);
   case CL_DEVICE_MAX_ON_DEVICE_EVENTS:
     POCL_RETURN_GETINFO(cl_uint, device->max_events);
   case CL_DEVICE_MAX_ON_DEVICE_QUEUES:
@@ -309,25 +295,83 @@ POname(clGetDeviceInfo)(cl_device_id   device,
     POCL_RETURN_GETINFO(cl_uint, 0);
   case CL_DEVICE_PREFERRED_PLATFORM_ATOMIC_ALIGNMENT:
     POCL_RETURN_GETINFO(cl_uint, 0);
-  case CL_DEVICE_SPIR_VERSIONS:
-    if (strstr (device->extensions, "cl_khr_spir"))
-      POCL_RETURN_GETINFO_STR ("1.2");
-    else
-      POCL_RETURN_GETINFO_STR ("");
+
   case CL_DEVICE_QUEUE_ON_DEVICE_PROPERTIES:
     POCL_RETURN_GETINFO(cl_command_queue_properties, device->on_dev_queue_props);
   case CL_DEVICE_QUEUE_ON_HOST_PROPERTIES:
     POCL_RETURN_GETINFO(cl_command_queue_properties, device->on_host_queue_props);
-
   case CL_DEVICE_GLOBAL_VARIABLE_PREFERRED_TOTAL_SIZE:
     POCL_RETURN_GETINFO(size_t, device->global_var_pref_size);
   case CL_DEVICE_MAX_GLOBAL_VARIABLE_SIZE:
     POCL_RETURN_GETINFO(size_t, device->global_var_max_size);
-  case CL_DEVICE_IL_VERSION:
-    if (device->spirv_version)
-      POCL_RETURN_GETINFO_STR (device->spirv_version);
+  case CL_DEVICE_SPIR_VERSIONS:
+    if (device->supported_spir_versions)
+      POCL_RETURN_GETINFO_STR (device->supported_spir_versions);
     else
       POCL_RETURN_GETINFO_STR ("");
+  case CL_DEVICE_IL_VERSION:
+    if (device->supported_spir_v_versions)
+      POCL_RETURN_GETINFO_STR (device->supported_spir_v_versions);
+    else
+      POCL_RETURN_GETINFO_STR ("");
+  case CL_DEVICE_MAX_NUM_SUB_GROUPS:
+    POCL_RETURN_GETINFO (cl_uint, device->max_num_sub_groups);
+  case CL_DEVICE_SUB_GROUP_INDEPENDENT_FORWARD_PROGRESS:
+    POCL_RETURN_GETINFO (cl_bool,
+                         device->sub_group_independent_forward_progress);
+
+  /** OpenCL 3.0 queries **/
+
+  case CL_DEVICE_ATOMIC_MEMORY_CAPABILITIES:
+    POCL_RETURN_GETINFO (cl_device_atomic_capabilities,
+                         device->atomic_memory_capabilities);
+  case CL_DEVICE_ATOMIC_FENCE_CAPABILITIES:
+    POCL_RETURN_GETINFO (cl_device_atomic_capabilities,
+                         device->atomic_fence_capabilities);
+  case CL_DEVICE_NON_UNIFORM_WORK_GROUP_SUPPORT:
+    POCL_RETURN_GETINFO (cl_bool, device->non_uniform_work_group_support);
+  case CL_DEVICE_WORK_GROUP_COLLECTIVE_FUNCTIONS_SUPPORT:
+    POCL_RETURN_GETINFO (cl_bool, CL_FALSE);
+  case CL_DEVICE_GENERIC_ADDRESS_SPACE_SUPPORT:
+    POCL_RETURN_GETINFO (cl_bool, CL_FALSE);
+  case CL_DEVICE_DEVICE_ENQUEUE_CAPABILITIES:
+    POCL_RETURN_GETINFO (cl_uint, 0);
+  case CL_DEVICE_PIPE_SUPPORT:
+    POCL_RETURN_GETINFO (cl_bool, device->pipe_support);
+  case CL_DEVICE_PREFERRED_WORK_GROUP_SIZE_MULTIPLE:
+    POCL_RETURN_GETINFO (size_t, device->preferred_wg_size_multiple);
+  case CL_DEVICE_NUMERIC_VERSION:
+    POCL_RETURN_GETINFO (cl_version, device->version_as_cl);
+
+  case CL_DEVICE_LATEST_CONFORMANCE_VERSION_PASSED:
+    if (device->version_of_latest_passed_cts)
+      POCL_RETURN_GETINFO_STR (device->version_of_latest_passed_cts);
+    else
+      POCL_RETURN_GETINFO_STR ("");
+
+  case CL_DEVICE_EXTENSIONS_WITH_VERSION:
+    POCL_RETURN_GETINFO_ARRAY (cl_name_version,
+                               device->num_extensions_with_version,
+                               device->extensions_with_version);
+
+  case CL_DEVICE_ILS_WITH_VERSION:
+    POCL_RETURN_GETINFO_ARRAY (cl_name_version, device->num_ils_with_version,
+                               device->ils_with_version);
+
+  case CL_DEVICE_BUILT_IN_KERNELS_WITH_VERSION:
+    POCL_RETURN_GETINFO_ARRAY (cl_name_version,
+                               device->num_builtin_kernels_with_version,
+                               device->builtin_kernels_with_version);
+
+  case CL_DEVICE_OPENCL_C_ALL_VERSIONS:
+    POCL_RETURN_GETINFO_ARRAY (cl_name_version,
+                               device->num_opencl_c_with_version,
+                               device->opencl_c_with_version);
+
+  case CL_DEVICE_OPENCL_C_FEATURES:
+    POCL_RETURN_GETINFO_ARRAY (cl_name_version,
+                               device->num_opencl_features_with_version,
+                               device->opencl_features_with_version);
   }
 
   if(device->ops->get_device_info_ext != NULL) {
