@@ -28,22 +28,22 @@
 CL_API_ENTRY void * CL_API_CALL
 POname (clGetExtensionFunctionAddressForPlatform) (cl_platform_id  platform,
                                                    const char *func_name)
-CL_EXT_SUFFIX__VERSION_1_2
+CL_API_SUFFIX__VERSION_1_2
 {
   cl_platform_id pocl_platform;
   cl_uint actual_num = 0;
   POname (clGetPlatformIDs) (1, &pocl_platform, &actual_num);
   if (actual_num != 1)
     {
-      POCL_MSG_WARN ("Couldn't get the platform ID of Pocl platform\n");
+      POCL_MSG_WARN ("Couldn't get the platform ID of PoCL platform\n");
       return NULL;
     }
 
   assert (pocl_platform);
   if (platform != pocl_platform)
     {
-      POCL_MSG_PRINT_INFO ("Requested Function Address not "
-                           "for Pocl platform, ignoring\n");
+      POCL_MSG_WARN ("Requested Function Address not "
+                     "for PoCL platform, ignoring\n");
       return NULL;
     }
 
@@ -52,6 +52,11 @@ CL_EXT_SUFFIX__VERSION_1_2
     return (void *)&POname(clIcdGetPlatformIDsKHR);
 #endif
 
+#ifdef BUILD_PROXY
+  if (strcmp (func_name, "clGetGLContextInfoKHR") == 0)
+    return (void *)&POname (clGetGLContextInfoKHR);
+#endif
+
   if (strcmp (func_name, "clSetContentSizeBufferPoCL") == 0)
     return (void *)&POname (clSetContentSizeBufferPoCL);
 
