@@ -23,8 +23,13 @@
 
 #include "pocl_util.h"
 
+#define S(A) #A
+#define STRINGIFY(X) S(X)
+
 static const char *pocl_version
-    = "OpenCL " POCL_CL_VERSION " pocl " POCL_VERSION_FULL
+    = "OpenCL " STRINGIFY(POCL_PLATFORM_VERSION_MAJOR)
+      "." STRINGIFY(POCL_PLATFORM_VERSION_MINOR)
+      " PoCL " POCL_VERSION_FULL
 
 #if defined(_WIN32)
    #if defined(_WIN64)
@@ -95,6 +100,10 @@ static const char *pocl_version
       ", UBSAN"
 #endif
 
+#ifdef BUILD_PROXY
+      ", PROXY"
+#endif
+
 #ifdef BUILD_CUDA
       ", CUDA"
 #endif
@@ -116,6 +125,49 @@ static const char *pocl_version
 #endif
     ;
 
+static const cl_name_version pocl_platform_extensions[] = {
+#ifdef BUILD_ICD
+  { CL_MAKE_VERSION (1, 0, 0), "cl_khr_icd" },
+#endif
+  { CL_MAKE_VERSION (1, 0, 0), "cl_pocl_content_size" }
+};
+static const size_t pocl_platform_extensions_num
+    = sizeof (pocl_platform_extensions) / sizeof (cl_name_version);
+
+static const cl_version pocl_numeric_version
+    = CL_MAKE_VERSION (POCL_PLATFORM_VERSION_MAJOR,
+                       POCL_PLATFORM_VERSION_MINOR,
+                       POCL_PLATFORM_VERSION_PATCH);
+
+static void
+pocl_cl_name_version_to_str (char *output, size_t limit)
+{
+  output[0] = 0;
+  if (pocl_platform_extensions_num == 0)
+    {
+      return;
+    }
+  char *str_ptr = output;
+  size_t i, remain = limit;
+  for (i = 0; i < pocl_platform_extensions_num; ++i)
+    {
+      // space + NULL
+      size_t len = strlen (pocl_platform_extensions[i].name);
+      if (len + 2 > remain)
+        break;
+
+      if (i > 0)
+        {
+          strcpy (str_ptr, " ");
+          str_ptr++;
+          remain--;
+        }
+      strcat (str_ptr, pocl_platform_extensions[i].name);
+      str_ptr += len;
+      remain -= len;
+    }
+}
+
 CL_API_ENTRY cl_int CL_API_CALL
 POname(clGetPlatformInfo)(cl_platform_id   platform,
                   cl_platform_info param_name,
@@ -139,9 +191,13 @@ POname(clGetPlatformInfo)(cl_platform_id   platform,
       POCL_RETURN_GETINFO_STR ("FULL_PROFILE");
 
     case CL_PLATFORM_VERSION:
-
       POCL_RETURN_GETINFO_STR (pocl_version);
 
+    case CL_PLATFORM_NUMERIC_VERSION:
+      // Let's asume that only single digit is used for major and minor
+      // versioning , and there is no patch.
+      POCL_RETURN_GETINFO (cl_version, pocl_numeric_version);
+
     case CL_PLATFORM_NAME:
       POCL_RETURN_GETINFO_STR("Portable Computing Language");
 
@@ -149,15 +205,23 @@ POname(clGetPlatformInfo)(cl_platform_id   platform,
       POCL_RETURN_GETINFO_STR("The pocl project");
 
     case CL_PLATFORM_EXTENSIONS:
-      POCL_RETURN_GETINFO_STR (
-#ifdef BUILD_ICD
-          "cl_khr_icd"
-#endif
-          " cl_pocl_content_size");
+      {
+        char result[2048];
+        pocl_cl_name_version_to_str (result, 2048);
+        POCL_RETURN_GETINFO_STR (result);
+      }
+
+    case CL_PLATFORM_EXTENSIONS_WITH_VERSION:
+      POCL_RETURN_GETINFO_ARRAY (cl_name_version,
+                                 pocl_platform_extensions_num,
+                                 pocl_platform_extensions);
 
     case CL_PLATFORM_ICD_SUFFIX_KHR:
       POCL_RETURN_GETINFO_STR("POCL");
 
+    case CL_PLATFORM_HOST_TIMER_RESOLUTION:
+      POCL_RETURN_GETINFO(cl_ulong, 0);
+
     default:
       return CL_INVALID_VALUE;
   }
