@@ -23,8 +23,7 @@
 */
 
 #include "devices/devices.h"
-#include "pocl_cl.h"
-#include "pocl_util.h"
+#include "pocl_runtime_config.h"
 
 #ifdef ENABLE_LLVM
 #include "pocl_llvm.h"
@@ -55,22 +54,23 @@ POname(clReleaseContext)(cl_context context) CL_API_SUFFIX__VERSION_1_0
 
       POCL_MSG_PRINT_REFCOUNTS ("Free Context %p\n", context);
 
-      if (context->num_devices == 0)
-        goto FINISH;
-
       /* The context holds references to all its devices,
          memory objects, command-queues etc. Release the
          references and let the objects to get freed. */
       unsigned i;
       for (i = 0; i < context->num_devices; ++i)
         {
+          cl_device_id dev = context->devices[i];
           if (context->default_queues && context->default_queues[i])
             POname (clReleaseCommandQueue) (context->default_queues[i]);
-          POname(clReleaseDevice) (context->devices[i]);
+          if (dev->ops->free_context)
+            dev->ops->free_context (dev, context);
+          POname (clReleaseDevice) (context->devices[i]);
         }
+
       POCL_MEM_FREE (context->default_queues);
-      POCL_MEM_FREE(context->devices);
-      POCL_MEM_FREE(context->properties);
+      POCL_MEM_FREE (context->devices);
+      POCL_MEM_FREE (context->properties);
 
       for (i = 0; i < NUM_OPENCL_IMAGE_TYPES; ++i)
         POCL_MEM_FREE (context->image_formats[i]);
@@ -79,6 +79,17 @@ POname(clReleaseContext)(cl_context context) CL_API_SUFFIX__VERSION_1_0
       pocl_llvm_release_context (context);
 #endif
 
+      /* Fire any registered destructor callbacks */
+      context_destructor_callback_t *next_callback,
+          *callback = context->destructor_callbacks;
+      while (callback)
+        {
+          callback->pfn_notify (context, callback->user_data);
+          next_callback = callback->next;
+          free (callback);
+          callback = next_callback;
+        }
+
       POCL_DESTROY_OBJECT (context);
       POCL_MEM_FREE(context);
 
@@ -90,7 +101,6 @@ POname(clReleaseContext)(cl_context context) CL_API_SUFFIX__VERSION_1_0
       VG_REFC_NONZERO (context);
     }
 
-FINISH:
   POCL_UNLOCK (pocl_context_handling_lock);
 
   return CL_SUCCESS;
@@ -119,6 +129,8 @@ pocl_check_uninit_devices ()
 
   usleep (100000);
 
+  pocl_event_tracing_finish ();
+
   POCL_LOCK (pocl_context_handling_lock);
   if (cl_context_count == 0)
     {
