@@ -26,7 +26,7 @@
 CL_API_ENTRY cl_int CL_API_CALL
 POname(clRetainEvent)(cl_event  event ) CL_API_SUFFIX__VERSION_1_0
 {
-  POCL_RETURN_ERROR_COND((event == NULL), CL_INVALID_EVENT);
+  POCL_RETURN_ERROR_COND ((!IS_CL_OBJECT_VALID (event)), CL_INVALID_EVENT);
 
   int refc;
   POCL_RETAIN_OBJECT_REFCOUNT (event, refc);
