@@ -384,7 +384,7 @@ cl_int pocl_accel_alloc_mem_obj(cl_device_id device, cl_mem mem_obj,
   if ((mem_obj->flags & CL_MEM_ALLOC_HOST_PTR) && (mem_obj->mem_host_ptr == NULL))
     return CL_MEM_OBJECT_ALLOCATION_FAILURE;
 
-  chunk = alloc_buffer_from_region(&data->AllocRegion, mem_obj->size);
+  chunk = pocl_alloc_buffer_from_region(&data->AllocRegion, mem_obj->size);
   if (chunk == NULL)
     return CL_MEM_OBJECT_ALLOCATION_FAILURE;
 
@@ -410,7 +410,7 @@ void pocl_accel_free(cl_device_id device, cl_mem mem) {
                           mem->size, chunk->start_address);
 
   assert(chunk != NULL);
-  free_chunk (chunk);
+  pocl_free_chunk(chunk);
 
   p->mem_ptr = NULL;
   p->version = 0;
@@ -521,7 +521,8 @@ cl_int pocl_accel_init(unsigned j, cl_device_id dev, const char *parameters) {
   D->DataMemory.Map(D->BaseAddress + 2 * max_region, dmem_size, mem_fd);
   D->ParameterMemory.Map(D->BaseAddress + 3 * max_region, pmem_size, mem_fd);
 
-  init_mem_region(&D->AllocRegion, D->ParameterMemory.PhysAddress, pmem_size);
+  pocl_init_mem_region(&D->AllocRegion, D->ParameterMemory.PhysAddress,
+                       pmem_size);
 
   // memory mapping done
   close(mem_fd);
@@ -567,13 +568,7 @@ unsigned int pocl_accel_probe(struct pocl_device_ops *ops) {
 
 char *pocl_accel_build_hash(cl_device_id /*device*/) {
   char *res = (char *)calloc(1000, sizeof(char));
-#ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-  char *name = get_llvm_cpu_name();
-  snprintf(res, 1000, "accel-%s-%s", HOST_DEVICE_BUILD_HASH, name);
-  POCL_MEM_FREE(name);
-#else
   snprintf(res, 1000, "accel-%s", HOST_DEVICE_BUILD_HASH);
-#endif
   return res;
 }
 
@@ -703,7 +698,8 @@ size_t scheduleNDRange(AccelData *data, _cl_command_run *run, size_t arg_size,
   }
   // Additional space for a signal
   size_t extraAlloc = sizeof(uint32_t);
-  chunk_info_t *chunk = alloc_buffer_from_region(&data->AllocRegion, arg_size + extraAlloc);
+  chunk_info_t *chunk =
+      pocl_alloc_buffer_from_region(&data->AllocRegion, arg_size + extraAlloc);
   assert(chunk && "Failed to allocate signal/argument buffer");
 
   POCL_MSG_PRINT_INFO("accel: allocated 0x%zx bytes for signal/arguments "
@@ -776,6 +772,9 @@ void pocl_accel_run(void *data, _cl_command_node *cmd) {
   pocl_kernel_metadata_t *meta = kernel->meta;
   struct pocl_context *pc = &cmd->command.run.pc;
 
+  if (pc->num_groups[0] == 0 || pc->num_groups[1] == 0 || pc->num_groups[2] == 0)
+    return;
+
   // First pass to figure out total argument size
   size_t arg_size = 0;
   for (i = 0; i < meta->num_args; ++i) {
