@@ -1,7 +1,7 @@
 /* basic.c - a minimalistic single core pocl device driver layer implementation
 
    Copyright (c) 2011-2013 Universidad Rey Juan Carlos and
-                 2011-2020 Pekka Jääskeläinen
+                 2011-2021 Pekka Jääskeläinen
 
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
@@ -77,8 +77,8 @@ pocl_basic_init_device_ops(struct pocl_device_ops *ops)
   ops->reinit = pocl_basic_reinit;
   ops->init = pocl_basic_init;
 
-  ops->alloc_mem_obj = pocl_basic_alloc_mem_obj;
-  ops->free = pocl_basic_free;
+  ops->alloc_mem_obj = pocl_driver_alloc_mem_obj;
+  ops->free = pocl_driver_free;
 
   ops->read = pocl_driver_read;
   ops->read_rect = pocl_driver_read_rect;
@@ -125,7 +125,7 @@ pocl_basic_init_device_ops(struct pocl_device_ops *ops)
   ops->svm_map = NULL;
   ops->svm_unmap = NULL;
   ops->svm_copy = pocl_basic_svm_copy;
-  ops->svm_fill = pocl_basic_svm_fill;
+  ops->svm_fill = pocl_driver_svm_fill;
 
   ops->create_kernel = NULL;
   ops->free_kernel = NULL;
@@ -144,7 +144,7 @@ pocl_basic_build_hash (cl_device_id device)
 {
   char* res = calloc(1000, sizeof(char));
 #ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-  char *name = get_llvm_cpu_name ();
+  char *name = pocl_get_llvm_cpu_name ();
   snprintf (res, 1000, "basic-%s-%s", HOST_DEVICE_BUILD_HASH, name);
   POCL_MEM_FREE (name);
 #else
@@ -190,27 +190,53 @@ pocl_basic_init (unsigned j, cl_device_id device, const char* parameters)
   pocl_init_default_device_infos (device);
   /* 0 is the host memory shared with all drivers that use it */
   device->global_mem_id = 0;
+
+  device->version_of_latest_passed_cts = HOST_DEVICE_LATEST_CTS_PASS;
   device->extensions = HOST_DEVICE_EXTENSIONS;
 
+#if (HOST_DEVICE_CL_VERSION_MAJOR >= 3)
+  device->features = HOST_DEVICE_FEATURES_30;
+
+  pocl_setup_opencl_c_with_version (device, CL_TRUE);
+  pocl_setup_features_with_version (device);
+#else
+  pocl_setup_opencl_c_with_version (device, CL_FALSE);
+#endif
+
+  pocl_setup_extensions_with_version (device);
+
+  /* builtin kernels.. skip, basic/pthread doesn't have any
+  pocl_setup_builtin_kernels_with_version (device); */
+
+  pocl_setup_ils_with_version (device);
+
+  device->on_host_queue_props
+      = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_PROFILING_ENABLE;
+
+#if (!defined(ENABLE_CONFORMANCE)                                             \
+     || (defined(ENABLE_CONFORMANCE) && (HOST_DEVICE_CL_VERSION_MAJOR >= 3)))
   /* full memory consistency model for atomic memory and fence operations
-  except CL_DEVICE_ATOMIC_SCOPE_ALL_DEVICES. see 
-  https://www.khronos.org/registry/OpenCL/specs/3.0-unified/html/OpenCL_API.html#opencl-3.0-backwards-compatibility*/
+  https://www.khronos.org/registry/OpenCL/specs/3.0-unified/html/OpenCL_API.html#opencl-3.0-backwards-compatibility */
   device->atomic_memory_capabilities = CL_DEVICE_ATOMIC_ORDER_RELAXED
                                        | CL_DEVICE_ATOMIC_ORDER_ACQ_REL
                                        | CL_DEVICE_ATOMIC_ORDER_SEQ_CST
-                                       | CL_DEVICE_ATOMIC_SCOPE_WORK_GROUP 
-                                       | CL_DEVICE_ATOMIC_SCOPE_DEVICE;
+                                       | CL_DEVICE_ATOMIC_SCOPE_WORK_GROUP
+                                       | CL_DEVICE_ATOMIC_SCOPE_DEVICE
+                                       | CL_DEVICE_ATOMIC_SCOPE_ALL_DEVICES;
   device->atomic_fence_capabilities = CL_DEVICE_ATOMIC_ORDER_RELAXED
-                                      | CL_DEVICE_ATOMIC_ORDER_ACQ_REL
-                                      | CL_DEVICE_ATOMIC_ORDER_SEQ_CST
-                                      | CL_DEVICE_ATOMIC_SCOPE_WORK_ITEM 
-                                      | CL_DEVICE_ATOMIC_SCOPE_WORK_GROUP 
-                                      | CL_DEVICE_ATOMIC_SCOPE_DEVICE;
+                                       | CL_DEVICE_ATOMIC_ORDER_ACQ_REL
+                                       | CL_DEVICE_ATOMIC_ORDER_SEQ_CST
+                                       | CL_DEVICE_ATOMIC_SCOPE_WORK_ITEM
+                                       | CL_DEVICE_ATOMIC_SCOPE_WORK_GROUP
+                                       | CL_DEVICE_ATOMIC_SCOPE_DEVICE;
 
   device->svm_allocation_priority = 1;
+
   /* OpenCL 2.0 properties */
   device->svm_caps = CL_DEVICE_SVM_COARSE_GRAIN_BUFFER
-                     | CL_DEVICE_SVM_FINE_GRAIN_BUFFER | CL_DEVICE_SVM_ATOMICS;
+                     | CL_DEVICE_SVM_FINE_GRAIN_BUFFER
+                     | CL_DEVICE_SVM_ATOMICS;
+#endif
 
   /* hwloc probes OpenCL device info at its initialization in case
      the OpenCL extension is enabled. This causes to printout 
@@ -244,49 +270,6 @@ pocl_basic_init (unsigned j, cl_device_id device, const char* parameters)
   return ret;
 }
 
-
-cl_int
-pocl_basic_alloc_mem_obj (cl_device_id device, cl_mem mem, void* host_ptr)
-{
-  pocl_mem_identifier *p = &mem->device_ptrs[device->global_mem_id];
-
-  /* let other drivers preallocate */
-  if ((mem->flags & CL_MEM_ALLOC_HOST_PTR) && (mem->mem_host_ptr == NULL))
-    return CL_MEM_OBJECT_ALLOCATION_FAILURE;
-
-  /* malloc mem_host_ptr then increase refcount */
-  pocl_alloc_or_retain_mem_host_ptr (mem);
-
-  cl_device_id svm_dev = mem->context->svm_allocdev;
-  /* if we have a device which shares global memory with host,
-   * and it needs to do anything to make allocations accessible
-   * to itself, do it here */
-  if (svm_dev && svm_dev->global_mem_id == 0 && svm_dev->ops->svm_register)
-    svm_dev->ops->svm_register (svm_dev, mem->mem_host_ptr, mem->size);
-
-  p->version = mem->mem_host_ptr_version;
-  p->mem_ptr = mem->mem_host_ptr;
-
-  POCL_MSG_PRINT_MEMORY ("Basic device ALLOC %p / size %zu \n", p->mem_ptr,
-                         mem->size);
-
-  return CL_SUCCESS;
-}
-
-
-void
-pocl_basic_free (cl_device_id device, cl_mem mem)
-{
-  cl_device_id svm_dev = mem->context->svm_allocdev;
-  if (svm_dev && svm_dev->global_mem_id == 0 && svm_dev->ops->svm_unregister)
-    svm_dev->ops->svm_unregister (svm_dev, mem->mem_host_ptr, mem->size);
-
-  pocl_mem_identifier *p = &mem->device_ptrs[device->global_mem_id];
-  pocl_release_mem_host_ptr (mem);
-  p->mem_ptr = NULL;
-  p->version = 0;
-}
-
 void
 pocl_basic_run (void *data, _cl_command_node *cmd)
 {
@@ -298,6 +281,9 @@ pocl_basic_run (void *data, _cl_command_node *cmd)
   pocl_kernel_metadata_t *meta = kernel->meta;
   struct pocl_context *pc = &cmd->command.run.pc;
 
+  if (pc->num_groups[0] == 0 || pc->num_groups[1] == 0 || pc->num_groups[2] == 0)
+    return;
+
   assert (data != NULL);
   d = (struct data *) data;
 
@@ -830,12 +816,3 @@ pocl_basic_svm_copy (cl_device_id dev, void *__restrict__ dst,
 {
   memcpy (dst, src, size);
 }
-
-void
-pocl_basic_svm_fill (cl_device_id dev, void *__restrict__ svm_ptr, size_t size,
-                     void *__restrict__ pattern, size_t pattern_size)
-{
-  pocl_mem_identifier temp;
-  temp.mem_ptr = svm_ptr;
-  pocl_driver_memfill (dev->data, &temp, NULL, size, 0, pattern, pattern_size);
-}
