@@ -2,7 +2,7 @@
               implementations
 
    Copyright (c) 2011-2013 Universidad Rey Juan Carlos
-                 2011-2019 Pekka Jääskeläinen
+                 2011-2021 Pekka Jääskeläinen
 
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
@@ -29,6 +29,7 @@
 #define _POSIX_C_SOURCE 200809L
 
 #include <assert.h>
+#include <ctype.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -151,7 +152,8 @@ llvm_codegen (char *output, unsigned device_i, cl_kernel kernel,
   else
     {
       char kernel_parallel_path[POCL_FILENAME_LENGTH];
-      pocl_cache_kernel_cachedir_path (kernel_parallel_path, program, device_i,
+      pocl_cache_kernel_cachedir_path (kernel_parallel_path, program,
+                                       command->program_device_i,
                                        kernel, "", command, specialize);
       error = pocl_mkdir_p (kernel_parallel_path);
     }
@@ -292,18 +294,6 @@ pocl_fill_dev_image_t (dev_image_t *di, struct pocl_argument *parg,
   di->_data = (mem->device_ptrs[device->global_mem_id].mem_ptr);
 }
 
-void
-pocl_ndrange_node_cleanup(_cl_command_node *node)
-{
-  cl_uint i;
-  for (i = 0; i < node->command.run.kernel->meta->num_args; ++i)
-    {
-      pocl_aligned_free (node->command.run.arguments[i].value);
-    }
-  POCL_MEM_FREE (node->command.run.arguments);
-  POname(clReleaseKernel)(node->command.run.kernel);
-}
-
 
 /**
  * executes given command. Call with node->event UNLOCKED.
@@ -388,7 +378,6 @@ pocl_exec_command (_cl_command_node *node)
          cmd->memfill.pattern,
          cmd->memfill.pattern_size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event Fill Buffer           ");
-      pocl_aligned_free (cmd->memfill.pattern);
       break;
 
     case CL_COMMAND_READ_BUFFER_RECT:
@@ -611,7 +600,6 @@ pocl_exec_command (_cl_command_node *node)
           dev->ops->unmap_image (dev->data, cmd->unmap.mem_id, mem,
                                  cmd->unmap.mapping);
         }
-      pocl_unmap_command_finished (dev, mem_id, mem, cmd->unmap.mapping);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Unmap Mem obj         ");
       break;
 
@@ -620,14 +608,12 @@ pocl_exec_command (_cl_command_node *node)
       assert (dev->ops->run);
       dev->ops->run (dev->data, node);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event Enqueue NDRange       ");
-      pocl_ndrange_node_cleanup(node);
       break;
 
     case CL_COMMAND_NATIVE_KERNEL:
       pocl_update_event_running (event);
       assert (dev->ops->run_native);
       dev->ops->run_native (dev->data, node);
-      POCL_MEM_FREE (node->command.native.args);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event Native Kernel         ");
       break;
 
@@ -651,7 +637,24 @@ pocl_exec_command (_cl_command_node *node)
            cmd->svm_free.data);
       else
         for (i = 0; i < cmd->svm_free.num_svm_pointers; i++)
-          dev->ops->svm_free (dev, cmd->svm_free.svm_pointers[i]);
+          {
+            void *ptr = cmd->svm_free.svm_pointers[i];
+            POCL_LOCK_OBJ (event->context);
+            pocl_svm_ptr *tmp = NULL, *item = NULL;
+            DL_FOREACH_SAFE (event->context->svm_ptrs, item, tmp)
+            {
+              if (item->svm_ptr == ptr)
+                {
+                  DL_DELETE (event->context->svm_ptrs, item);
+                  break;
+                }
+            }
+            POCL_UNLOCK_OBJ (event->context);
+            assert (item);
+            POCL_MEM_FREE (item);
+            POname (clReleaseContext) (event->context);
+            dev->ops->svm_free (dev, ptr);
+          }
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event SVM Free              ");
       break;
 
@@ -698,15 +701,21 @@ pocl_exec_command (_cl_command_node *node)
                           cmd->svm_fill.pattern,
                           cmd->svm_fill.pattern_size);
       POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event SVM MemFill           ");
-      pocl_aligned_free (cmd->svm_fill.pattern);
+      break;
+
+    case CL_COMMAND_SVM_MIGRATE_MEM:
+      pocl_update_event_running (event);
+      if (dev->ops->svm_migrate)
+        dev->ops->svm_migrate (dev, cmd->svm_migrate.num_svm_pointers,
+                               cmd->svm_migrate.svm_pointers,
+                               cmd->svm_migrate.sizes);
+      POCL_UPDATE_EVENT_COMPLETE_MSG (event, "Event SVM MemFill           ");
       break;
 
     default:
       POCL_ABORT_UNIMPLEMENTED("");
       break;
     }
-
-  pocl_mem_manager_free_command (node);
 }
 
 /* call with brc_event UNLOCKED. */
@@ -737,6 +746,16 @@ pocl_broadcast (cl_event brc_event)
                 target->event->command->device, target->event, brc_event);
           }
 
+        if (pocl_is_tracing_enabled() && target->event->meta_data)
+          {
+            pocl_event_md *md = target->event->meta_data;
+            for (size_t i = 0; i < md->num_deps; ++i)
+              if (md->dep_ids[i] == brc_event->id)
+                {
+                  md->dep_ts[i] = brc_event->time_end;
+                  break;
+                }
+          }
         LL_DELETE (brc_event->notify_list, target);
         pocl_unlock_events_inorder (brc_event, target->event);
         pocl_mem_manager_free_event_node (target);
@@ -778,6 +797,17 @@ pocl_fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg)
   }
 }
 
+/* Returns the width of the widest dimension in the grid of the given
+   run command. */
+size_t
+pocl_cmd_max_grid_dim_width (_cl_command_run *cmd)
+{
+  return max (max (cmd->pc.local_size[0] * cmd->pc.num_groups[0],
+                   cmd->pc.local_size[1] * cmd->pc.num_groups[1]),
+              cmd->pc.local_size[2] * cmd->pc.local_size[2]);
+}
+
+
 /* CPU driver stuff */
 
 #ifdef HAVE_DLFCN_H
@@ -898,7 +928,7 @@ pocl_check_kernel_disk_cache (_cl_command_node *command, int specialized)
   _cl_command_run *run_cmd = &command->command.run;
   cl_kernel k = run_cmd->kernel;
   cl_program p = k->program;
-  unsigned dev_i = command->device_i;
+  unsigned dev_i = command->program_device_i;
 
   /* First try to find a static WG binary for the local size as they
      are always more efficient than the dynamic ones.  Also, in case
@@ -923,7 +953,9 @@ pocl_check_kernel_disk_cache (_cl_command_node *command, int specialized)
       POCL_UNLOCK (pocl_llvm_codegen_lock);
       if (error)
         POCL_ABORT ("Final linking of kernel %s failed.\n", k->name);
-      POCL_MSG_PRINT_INFO ("Built a WG function: %s\n", module_fn);
+      POCL_MSG_PRINT_INFO ("Built a %sWG function: %s\n",
+                           specialized ? "specialized " : "generic ",
+                           module_fn);
       return module_fn;
 #else
       /* TODO: This should be caught earlier. */
@@ -938,12 +970,12 @@ pocl_check_kernel_disk_cache (_cl_command_node *command, int specialized)
       /* First try to find a specialized WG binary, if allowed by the
          command. */
       if (!run_cmd->force_generic_wg_func)
-        pocl_cache_final_binary_path (module_fn, p, dev_i, k, command, 0);
+        pocl_cache_final_binary_path (module_fn, p, dev_i, k, command, 1);
 
       if (run_cmd->force_generic_wg_func || !pocl_exists (module_fn))
         {
           /* Then check for a dynamic (non-specialized) kernel. */
-          pocl_cache_final_binary_path (module_fn, p, dev_i, k, command, 1);
+          pocl_cache_final_binary_path (module_fn, p, dev_i, k, command, 0);
           if (!pocl_exists (module_fn))
             POCL_ABORT ("Generic WG function binary does not exist.\n");
           POCL_MSG_PRINT_INFO ("Using a cached generic WG function: %s\n",
@@ -956,15 +988,6 @@ pocl_check_kernel_disk_cache (_cl_command_node *command, int specialized)
   return module_fn;
 }
 
-/* Returns the width of the widest dimension in the grid of the given
-   run command. */
-size_t
-pocl_cmd_max_grid_dim_width (_cl_command_run *cmd)
-{
-  return max (max (cmd->pc.local_size[0] * cmd->pc.num_groups[0],
-                   cmd->pc.local_size[1] * cmd->pc.num_groups[1]),
-              cmd->pc.local_size[2] * cmd->pc.local_size[2]);
-}
 
 /* Look for a dlhandle in the dlhandle cache for the given kernel command.
    If found, push the handle up in the cache to improve cache hit speed,
@@ -1018,6 +1041,11 @@ pocl_check_kernel_dlhandle_cache (_cl_command_node *command,
   const char *dl_error = NULL;
   _cl_command_run *run_cmd = &command->command.run;
 
+  /* Brute force mechanism to test relying on generic work-group functions
+     only. */
+  if (!pocl_get_bool_option("POCL_WORK_GROUP_SPECIALIZATION", 1))
+    specialize = 0;
+
   POCL_LOCK (pocl_dlhandle_lock);
   ci = fetch_dlhandle_cache_item (run_cmd, specialize);
   if (ci != NULL)
@@ -1084,6 +1112,7 @@ pocl_check_kernel_dlhandle_cache (_cl_command_node *command,
 
 #endif
 
+
 #define MIN_MAX_MEM_ALLOC_SIZE (128*1024*1024)
 
 /* accounting object for the main memory */
@@ -1280,7 +1309,6 @@ pocl_print_system_memory_stats()
                     system_memory.max_ever_allocated >> 10);
 }
 
-
 /* default WG size in each dimension & total WG size.
  * this should be reasonable for CPU */
 #define DEFAULT_WG_SIZE 4096
@@ -1289,6 +1317,27 @@ static const char *final_ld_flags[] =
   {"-lm", "-nostartfiles", HOST_LD_FLAGS_ARRAY, NULL};
 
 static cl_device_partition_property basic_partition_properties[1] = { 0 };
+
+#ifdef ENABLE_CONFORMANCE
+static const cl_image_format supported_image_formats[] = {
+  { CL_RGBA, CL_SNORM_INT8 },
+  { CL_RGBA, CL_SNORM_INT16 },
+  { CL_RGBA, CL_UNORM_INT8 },
+  { CL_RGBA, CL_UNORM_INT16 },
+  { CL_RGBA, CL_SIGNED_INT8 },
+  { CL_RGBA, CL_SIGNED_INT16 },
+  { CL_RGBA, CL_SIGNED_INT32 },
+  { CL_RGBA, CL_UNSIGNED_INT8 },
+  { CL_RGBA, CL_UNSIGNED_INT16 },
+  { CL_RGBA, CL_UNSIGNED_INT32 },
+  { CL_RGBA, CL_HALF_FLOAT },
+  { CL_RGBA, CL_FLOAT },
+  { CL_BGRA, CL_SNORM_INT8 },
+  { CL_BGRA, CL_UNORM_INT8 },
+  { CL_BGRA, CL_SIGNED_INT8 },
+  { CL_BGRA, CL_UNSIGNED_INT8 }
+};
+#else
 static const cl_image_format supported_image_formats[] = {
   { CL_A, CL_SNORM_INT8 },
   { CL_A, CL_SNORM_INT16 },
@@ -1300,7 +1349,20 @@ static const cl_image_format supported_image_formats[] = {
   { CL_A, CL_UNSIGNED_INT8 },
   { CL_A, CL_UNSIGNED_INT16 },
   { CL_A, CL_UNSIGNED_INT32 },
+  { CL_A, CL_HALF_FLOAT },
   { CL_A, CL_FLOAT },
+  { CL_R, CL_SNORM_INT8 },
+  { CL_R, CL_SNORM_INT16 },
+  { CL_R, CL_UNORM_INT8 },
+  { CL_R, CL_UNORM_INT16 },
+  { CL_R, CL_SIGNED_INT8 },
+  { CL_R, CL_SIGNED_INT16 },
+  { CL_R, CL_SIGNED_INT32 },
+  { CL_R, CL_UNSIGNED_INT8 },
+  { CL_R, CL_UNSIGNED_INT16 },
+  { CL_R, CL_UNSIGNED_INT32 },
+  { CL_R, CL_HALF_FLOAT },
+  { CL_R, CL_FLOAT },
   { CL_RGBA, CL_SNORM_INT8 },
   { CL_RGBA, CL_SNORM_INT16 },
   { CL_RGBA, CL_UNORM_INT8 },
@@ -1322,6 +1384,7 @@ static const cl_image_format supported_image_formats[] = {
   { CL_BGRA, CL_SIGNED_INT8 },
   { CL_BGRA, CL_UNSIGNED_INT8 }
 };
+#endif
 
 void
 pocl_init_default_device_infos (cl_device_id dev)
@@ -1389,8 +1452,12 @@ pocl_init_default_device_infos (cl_device_id dev)
   dev->image_support = CL_TRUE;
   /* Use the minimum values until we get a more sensible upper limit from
      somewhere. */
-  dev->max_read_image_args = dev->max_write_image_args
-      = dev->max_read_write_image_args = 128;
+  dev->max_read_image_args = dev->max_write_image_args = 128;
+#ifdef ENABLE_CONFORMANCE
+  dev->max_read_write_image_args = 0;
+#else
+  dev->max_read_write_image_args = 128;
+#endif
   dev->image2d_max_width = dev->image2d_max_height = 8192;
   dev->image3d_max_width = dev->image3d_max_height = dev->image3d_max_depth = 2048;
   dev->max_samplers = 16;
@@ -1453,6 +1520,7 @@ pocl_init_default_device_infos (cl_device_id dev)
   dev->linker_available = CL_TRUE;
   dev->spmd = CL_FALSE;
   dev->arg_buffer_launcher = CL_FALSE;
+  dev->grid_launcher = CL_FALSE;
   dev->workgroup_pass = CL_TRUE;
   dev->execution_capabilities = CL_EXEC_KERNEL | CL_EXEC_NATIVE_KERNEL;
   dev->platform = 0;
@@ -1489,19 +1557,29 @@ pocl_init_default_device_infos (cl_device_id dev)
   /* OpenCL 2.0 properties */
   dev->svm_caps = 0;
   /* TODO these are minimums, figure out whats a reasonable value */
-  dev->max_events = 1024;
-  dev->max_queues = 1;
-  dev->max_pipe_args = 16;
-  dev->max_pipe_active_res = 1;
-  dev->max_pipe_packet_size = 1024;
-  dev->dev_queue_pref_size = 16 * 1024;
-  dev->dev_queue_max_size = 256 * 1024;
-  dev->on_dev_queue_props = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE
-                               | CL_QUEUE_PROFILING_ENABLE;
+  dev->max_events = 0;
+  dev->max_queues = 0;
+
+  /* Default pipe support for PoCL devices */
+  dev->pipe_support = CL_FALSE;
+  /* Specification requires pipe values to be 0, when pipes are not supported
+   */
+  dev->max_pipe_args = 0;
+  dev->max_pipe_active_res = 0;
+  dev->max_pipe_packet_size = 0;
+
+  dev->dev_queue_pref_size = 0;
+  dev->dev_queue_max_size = 0;
+  dev->on_dev_queue_props = 0;
   dev->on_host_queue_props = CL_QUEUE_PROFILING_ENABLE;
   dev->has_64bit_long = 1;
   dev->autolocals_to_args = POCL_AUTOLOCALS_TO_ARGS_ALWAYS;
   dev->device_alloca_locals = 0;
+  dev->global_var_max_size = 0;
+  dev->global_var_pref_size = 0;
+  dev->non_uniform_work_group_support = CL_FALSE;
+  dev->max_num_sub_groups = 0;
+  dev->sub_group_independent_forward_progress = CL_FALSE;
 
 #ifdef ENABLE_LLVM
 
@@ -1509,15 +1587,20 @@ pocl_init_default_device_infos (cl_device_id dev)
 #ifdef HOST_CPU_FORCED
   dev->llvm_cpu = OCL_KERNEL_TARGET_CPU;
 #else
-  dev->llvm_cpu = get_llvm_cpu_name ();
+  dev->llvm_cpu = pocl_get_llvm_cpu_name ();
 #endif
 
-  dev->spirv_version = "SPIR-V_1.2";
 #else /* No compiler, no CPU info */
   dev->llvm_cpu = NULL;
   dev->llvm_target_triplet = "";
 #endif
 
+#ifdef ENABLE_SPIRV
+  dev->supported_spir_v_versions = "SPIR-V_1.2";
+#else
+  dev->supported_spir_v_versions = "";
+#endif
+
   /* OpenCL 3.0 properties */
   /* Minimum mandated capability */
   dev->atomic_memory_capabilities = CL_DEVICE_ATOMIC_ORDER_RELAXED
@@ -1526,3 +1609,230 @@ pocl_init_default_device_infos (cl_device_id dev)
                                     | CL_DEVICE_ATOMIC_ORDER_ACQ_REL
                                     | CL_DEVICE_ATOMIC_SCOPE_WORK_GROUP;
 }
+
+/*
+ * Input: a space-delim string, and an array of cl_name_version items
+ * (which should have all possible items that could occur in input string).
+ * Then for each item in string, finds its corresponding cl_name_version
+ * in the array, and copy it to the output array of cl_name_version items.
+ * Returns: number of items found
+ */
+
+static unsigned
+pocl_space_delim_string_to_cl_name_version_array (
+    cl_name_version **output, const char *input_str,
+    const cl_name_version *search_array, unsigned search_array_size)
+{
+  unsigned i, start = 0, end = 0, pos = 0, len = 0;
+  const char *p = input_str;
+  char item_name[256];
+  unsigned num_output = 0;
+  cl_name_version storage[256];
+
+  while (p[pos])
+    {
+      while (p[pos] && isspace ((unsigned char)p[pos]))
+        pos++;
+      start = pos;
+      if (p[pos] == 0)
+        break;
+      while (p[pos] && !isspace ((unsigned char)p[pos]))
+        pos++;
+      end = pos;
+      len = end - start;
+      if (len > 255)
+        {
+          POCL_MSG_WARN ("item too long: %u | %s", len, p + start);
+          continue;
+        }
+      memcpy (item_name, p + start, len);
+      item_name[len] = 0;
+
+      int found = -1;
+      for (i = 0; i < search_array_size; ++i)
+        {
+          if (strcmp (item_name, search_array[i].name) == 0)
+            {
+              found = i;
+              break;
+            }
+        }
+      if (found >= 0)
+        {
+          memcpy (&storage[num_output], &search_array[found],
+                  sizeof (cl_name_version));
+          ++num_output;
+          if (num_output == 256) {
+            POCL_MSG_WARN ("max items reached (256)\n");
+            break;
+          }
+        }
+      else
+        POCL_MSG_WARN ("could not find item: %s\n", item_name);
+    }
+
+  if (num_output)
+    {
+      cl_name_version *tmp = malloc (num_output * sizeof (cl_name_version));
+      memcpy (tmp, storage, (num_output * sizeof (cl_name_version)));
+      *output = tmp;
+    }
+  else
+    *output = NULL;
+
+  return num_output;
+}
+
+static const cl_name_version OPENCL_C_VERSIONS[]
+    = { { CL_MAKE_VERSION (1, 0, 0), "OpenCL C" },
+        { CL_MAKE_VERSION (1, 1, 0), "OpenCL C" },
+        { CL_MAKE_VERSION (1, 2, 0), "OpenCL C" },
+        { CL_MAKE_VERSION (3, 0, 0), "OpenCL C" } };
+
+void
+pocl_setup_opencl_c_with_version (cl_device_id dev, int supports_30)
+{
+  dev->opencl_c_with_version = OPENCL_C_VERSIONS;
+  dev->num_opencl_c_with_version = supports_30 ? 4 : 3;
+}
+
+static const cl_name_version OPENCL_EXTENSIONS[]
+    = { { CL_MAKE_VERSION (1, 0, 0), "cl_khr_byte_addressable_store" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_global_int32_base_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_global_int32_extended_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_local_int32_base_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_local_int32_extended_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_int64_base_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_int64_extended_atomics" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_3d_image_writes" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_fp16" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_fp64" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_nv_device_attribute_query" },
+
+        { CL_MAKE_VERSION (2, 0, 0), "cl_khr_depth_images" },
+        { CL_MAKE_VERSION (1, 0, 0), "cl_khr_image2d_from_buffer" },
+
+        { CL_MAKE_VERSION (2, 1, 0), "cl_khr_spir" },
+        { CL_MAKE_VERSION (2, 1, 0), "cl_khr_il_program" } };
+
+const size_t OPENCL_EXTENSIONS_NUM
+    = sizeof (OPENCL_EXTENSIONS) / sizeof (OPENCL_EXTENSIONS[0]);
+
+void
+pocl_setup_extensions_with_version (cl_device_id dev)
+{
+  cl_name_version *tmp = NULL;
+  unsigned ret = pocl_space_delim_string_to_cl_name_version_array (
+      &tmp, dev->extensions, OPENCL_EXTENSIONS, OPENCL_EXTENSIONS_NUM);
+
+  dev->num_extensions_with_version = ret;
+  dev->extensions_with_version = tmp;
+}
+
+static const cl_name_version OPENCL_SPIRV_VERSIONS[]
+    = { { CL_MAKE_VERSION (1, 0, 0), "SPIR-V" },
+        { CL_MAKE_VERSION (1, 1, 0), "SPIR-V" },
+        { CL_MAKE_VERSION (1, 2, 0), "SPIR-V" },
+        { CL_MAKE_VERSION (1, 3, 0), "SPIR-V" },
+        { CL_MAKE_VERSION (1, 4, 0), "SPIR-V" },
+        { CL_MAKE_VERSION (1, 5, 0), "SPIR-V" } };
+
+const size_t OPENCL_SPIRV_VERSIONS_NUM
+    = sizeof (OPENCL_SPIRV_VERSIONS) / sizeof (OPENCL_SPIRV_VERSIONS[0]);
+
+void
+pocl_setup_ils_with_version (cl_device_id dev)
+{
+  unsigned i, start = 0, end = 0, pos = 0, len = 0;
+  const char *p = dev->supported_spir_v_versions;
+  char item_name[256];
+  unsigned num_output = 0;
+  cl_name_version storage[256];
+
+  while (p[pos])
+    {
+      while (p[pos] && isspace ((unsigned char)p[pos]))
+        pos++;
+      start = pos;
+      if (p[pos] == 0)
+        break;
+      while (p[pos] && !isspace ((unsigned char)p[pos]))
+        pos++;
+      end = pos;
+      len = end - start;
+      if (len > 255)
+        {
+          POCL_MSG_WARN ("item too long: %u | %s", len, p + start);
+          continue;
+        }
+      memcpy (item_name, p + start, len);
+      item_name[len] = 0;
+
+      cl_version V = CL_MAKE_VERSION (0, 0, 0);
+      if (strncmp (item_name, "SPIR-V_1.", 9) == 0)
+        {
+          unsigned minor = item_name[9] - '0';
+          V = CL_MAKE_VERSION (1, minor, 0);
+        }
+      int found = -1;
+      for (i = 0; i < OPENCL_SPIRV_VERSIONS_NUM; ++i)
+        {
+          if (OPENCL_SPIRV_VERSIONS[i].version == V)
+            {
+              found = i;
+              break;
+            }
+        }
+      if (found >= 0)
+        {
+          memcpy (&storage[num_output], &OPENCL_SPIRV_VERSIONS[found],
+                  sizeof (cl_name_version));
+          ++num_output;
+          if (num_output >= 256)
+            break;
+        }
+      else
+        POCL_MSG_WARN ("could not find item: %s\n", item_name);
+    }
+
+  if (num_output)
+    {
+      cl_name_version *tmp = malloc (num_output * sizeof (cl_name_version));
+      memcpy (tmp, storage, (num_output * sizeof (cl_name_version)));
+      dev->ils_with_version = tmp;
+      dev->num_ils_with_version = num_output;
+    }
+}
+
+static const cl_name_version OPENCL_FEATURES[] = {
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_3d_image_writes" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_images" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_read_write_images" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_atomic_order_acq_rel" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_atomic_order_seq_cst" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_atomic_scope_device" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_atomic_scope_all_devices" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_fp16" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_fp64" },
+  { CL_MAKE_VERSION (3, 0, 0), "__opencl_c_int64" },
+};
+
+const size_t OPENCL_FEATURES_NUM
+    = sizeof (OPENCL_FEATURES) / sizeof (OPENCL_FEATURES[0]);
+
+void
+pocl_setup_features_with_version (cl_device_id dev)
+{
+  cl_name_version *tmp = NULL;
+  unsigned ret = pocl_space_delim_string_to_cl_name_version_array (
+      &tmp, dev->features, OPENCL_FEATURES, OPENCL_FEATURES_NUM);
+
+  dev->num_opencl_features_with_version = ret;
+  dev->opencl_features_with_version = tmp;
+}
+
+void
+pocl_setup_builtin_kernels_with_version (cl_device_id dev)
+{
+  /* implementation should use BIDescriptors */
+}
