@@ -32,10 +32,12 @@
 #include "pocl_image_types.h"
 #undef __CBUILD__
 
-#define XSETUP_DEVICE_CL_VERSION(A, B)             \
-  dev->cl_version_int = (A * 100) + (B * 10);     \
-  dev->cl_version_std = "CL" # A "." # B;         \
-  dev->version = "OpenCL " # A "." # B " pocl";
+#define XSETUP_DEVICE_CL_VERSION(A, B)                                        \
+  dev->version_as_int = (A * 100) + (B * 10);                                 \
+  dev->version_as_cl = CL_MAKE_VERSION (A, B, 0);                             \
+  dev->version = "OpenCL " #A "." #B " PoCL";                                 \
+  dev->opencl_c_version_as_opt = "CL" #A "." #B;                              \
+  dev->opencl_c_version_as_cl = CL_MAKE_VERSION (A, B, 0);
 
 #define SETUP_DEVICE_CL_VERSION(a, b) XSETUP_DEVICE_CL_VERSION(a, b)
 
@@ -70,9 +72,6 @@ void pocl_fill_dev_sampler_t (dev_sampler_t *ds, struct pocl_argument *parg);
 POCL_EXPORT
 void pocl_exec_command (_cl_command_node *node);
 
-POCL_EXPORT
-void pocl_ndrange_node_cleanup(_cl_command_node *node);
-
 POCL_EXPORT
 void pocl_broadcast (cl_event event);
 
@@ -82,6 +81,7 @@ void pocl_init_dlhandle_cache ();
 POCL_EXPORT
 char *pocl_check_kernel_disk_cache (_cl_command_node *cmd, int specialized);
 
+POCL_EXPORT
 size_t pocl_cmd_max_grid_dim_width (_cl_command_run *cmd);
 
 POCL_EXPORT
@@ -92,8 +92,10 @@ void pocl_check_kernel_dlhandle_cache (_cl_command_node *command,
 POCL_EXPORT
 void pocl_release_dlhandle_cache (_cl_command_node *cmd);
 
+POCL_EXPORT
 void pocl_setup_device_for_system_memory(cl_device_id device);
 
+POCL_EXPORT
 void pocl_reinit_system_memory();
 
 POCL_EXPORT
@@ -110,6 +112,20 @@ void pocl_print_system_memory_stats();
 POCL_EXPORT
 void pocl_init_default_device_infos (cl_device_id dev);
 
+POCL_EXPORT
+void pocl_setup_opencl_c_with_version (cl_device_id dev, int supports_30);
+
+POCL_EXPORT
+void pocl_setup_extensions_with_version (cl_device_id dev);
+
+POCL_EXPORT
+void pocl_setup_ils_with_version (cl_device_id dev);
+
+POCL_EXPORT
+void pocl_setup_features_with_version (cl_device_id dev);
+
+POCL_EXPORT
+void pocl_setup_builtin_kernels_with_version (cl_device_id dev);
 
 #ifdef __cplusplus
 }
