@@ -1,3 +1,27 @@
+/* common_driver.c - common code that can be reused between device driver
+   implementations
+
+   Copyright (c) 2011-2021 pocl developers
+
+   Permission is hereby granted, free of charge, to any person obtaining a copy
+   of this software and associated documentation files (the "Software"), to
+   deal in the Software without restriction, including without limitation the
+   rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+   sell copies of the Software, and to permit persons to whom the Software is
+   furnished to do so, subject to the following conditions:
+
+   The above copyright notice and this permission notice shall be included in
+   all copies or substantial portions of the Software.
+
+   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+   IN THE SOFTWARE.
+*/
+
 #include "config.h"
 
 #include <assert.h>
@@ -430,6 +454,57 @@ pocl_driver_free_mapping_ptr (void *data, pocl_mem_identifier *mem_id,
   return CL_SUCCESS;
 }
 
+cl_int
+pocl_driver_alloc_mem_obj (cl_device_id device, cl_mem mem, void *host_ptr)
+{
+  pocl_mem_identifier *p = &mem->device_ptrs[device->global_mem_id];
+
+  /* let other drivers preallocate */
+  if ((mem->flags & CL_MEM_ALLOC_HOST_PTR) && (mem->mem_host_ptr == NULL))
+    return CL_MEM_OBJECT_ALLOCATION_FAILURE;
+
+  /* malloc mem_host_ptr then increase refcount */
+  pocl_alloc_or_retain_mem_host_ptr (mem);
+
+  cl_device_id svm_dev = mem->context->svm_allocdev;
+  /* if we have a device which shares global memory with host,
+   * and it needs to do anything to make allocations accessible
+   * to itself, do it here */
+  if (svm_dev && svm_dev->global_mem_id == 0 && svm_dev->ops->svm_register)
+    svm_dev->ops->svm_register (svm_dev, mem->mem_host_ptr, mem->size);
+
+  p->version = mem->mem_host_ptr_version;
+  p->mem_ptr = mem->mem_host_ptr;
+
+  POCL_MSG_PRINT_MEMORY ("Basic device ALLOC %p / size %zu \n", p->mem_ptr,
+                         mem->size);
+
+  return CL_SUCCESS;
+}
+
+void
+pocl_driver_free (cl_device_id device, cl_mem mem)
+{
+  cl_device_id svm_dev = mem->context->svm_allocdev;
+  if (svm_dev && svm_dev->global_mem_id == 0 && svm_dev->ops->svm_unregister)
+    svm_dev->ops->svm_unregister (svm_dev, mem->mem_host_ptr, mem->size);
+
+  pocl_mem_identifier *p = &mem->device_ptrs[device->global_mem_id];
+  pocl_release_mem_host_ptr (mem);
+  p->mem_ptr = NULL;
+  p->version = 0;
+}
+
+void
+pocl_driver_svm_fill (cl_device_id dev, void *__restrict__ svm_ptr,
+                      size_t size, void *__restrict__ pattern,
+                      size_t pattern_size)
+{
+  pocl_mem_identifier temp;
+  temp.mem_ptr = svm_ptr;
+  pocl_driver_memfill (dev->data, &temp, NULL, size, 0, pattern, pattern_size);
+}
+
 /* These are implementations of compilation callbacks for all devices
  * that support compilation via LLVM. They take care of compilation/linking
  * of source/binary/spir down to parallel.bc level.
@@ -633,7 +708,7 @@ pocl_driver_setup_metadata (cl_device_id device, cl_program program,
 }
 
 int
-pocl_driver_supports_binary (cl_device_id device, const size_t length,
+pocl_driver_supports_binary (cl_device_id device, size_t length,
                              const char *binary)
 {
 #ifdef ENABLE_LLVM
@@ -689,7 +764,7 @@ pocl_driver_build_poclbinary (cl_program program, cl_uint device_i)
   assert (program->binaries[device_i]);
 
   cmd.device = device;
-  cmd.device_i = device_i;
+  cmd.program_device_i = device_i;
 
   struct _cl_kernel fake_k;
   memset (&fake_k, 0, sizeof (fake_k));
@@ -724,11 +799,65 @@ pocl_driver_build_poclbinary (cl_program program, cl_uint device_i)
       cmd.command.run.pc.global_offset[0] = cmd.command.run.pc.global_offset[1]
           = cmd.command.run.pc.global_offset[2] = 0;
 
-      /* Force generate a generic WG function to ensure generality. */
+      /* Force generate a generic WG function to ensure all local sizes
+         can be executed using the binary. */
       device->ops->compile_kernel (&cmd, kernel, device, 0);
-      /* Then generate a specialized one with goffset 0 since it's a very
-         common case. */
-      device->ops->compile_kernel (&cmd, kernel, device, 1);
+      /* Then generate specialized ones as requested via the
+         POCL_BINARY_SPECIALIZE_WG configuration option. */
+      char *temp
+          = strdup (pocl_get_string_option ("POCL_BINARY_SPECIALIZE_WG", ""));
+      char *token;
+      char *rest = temp;
+
+      while ((token = strtok_r (rest, ",", &rest)))
+        {
+          /* By default don't specialize for the origo global offset. */
+          cmd.command.run.pc.global_offset[0]
+              = cmd.command.run.pc.global_offset[1]
+              = cmd.command.run.pc.global_offset[2] = 1;
+
+          /* By default don't specialize for the local size. */
+          cmd.command.run.pc.local_size[0] = cmd.command.run.pc.local_size[1]
+              = cmd.command.run.pc.local_size[2] = 0;
+
+          /* By default don't specialize for a small grid size. */
+          cmd.command.run.force_large_grid_wg_func = 1;
+
+          /* The format of the specialization follows the format of the
+             cache directory. E.g. 128-1-1-goffs0, 13-1-1-goffs0-smallgrid
+             or 0-0-0-goffs0. We thus assume the local size is always given
+             first. */
+
+          char *param1 = NULL, *param2 = NULL;
+          int params_found
+              = sscanf (token, "%lu-%lu-%lu-%m[^-]-%m[^-]",
+                        &cmd.command.run.pc.local_size[0],
+                        &cmd.command.run.pc.local_size[1],
+                        &cmd.command.run.pc.local_size[2], &param1, &param2);
+          if (param1 != NULL)
+            {
+              if (strncmp (param1, "goffs0", 6) == 0)
+                {
+                  cmd.command.run.pc.global_offset[0]
+                      = cmd.command.run.pc.global_offset[1]
+                      = cmd.command.run.pc.global_offset[2] = 0;
+
+                  if (param2 != NULL && strncmp (param2, "smallgrid", 8) == 0)
+                    {
+                      cmd.command.run.force_large_grid_wg_func = 0;
+                    }
+                }
+              else if (strncmp (param1, "smallgrid", 8) == 0)
+                {
+                  cmd.command.run.force_large_grid_wg_func = 0;
+                }
+            }
+          free (param1);
+          free (param2);
+
+          device->ops->compile_kernel (&cmd, kernel, device, 1);
+        }
+      free (temp);
     }
 
   POCL_UNLOCK_OBJ (program);
