@@ -44,6 +44,11 @@
 #include <cuda.h>
 #include <cuda_runtime.h>
 
+#include <fcntl.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
+
 typedef struct pocl_cuda_device_data_s
 {
   CUdevice device;
@@ -52,6 +57,7 @@ typedef struct pocl_cuda_device_data_s
   cl_ulong epoch;
   char libdevice[PATH_MAX];
   pocl_lock_t compile_lock;
+  int supports_cu_mem_host_register;
 } pocl_cuda_device_data_t;
 
 typedef struct pocl_cuda_queue_data_s
@@ -286,12 +292,15 @@ pocl_cuda_init (unsigned j, cl_device_id dev, const char *parameters)
     ret = CL_INVALID_DEVICE;
 
   /* Get specific device name */
-  dev->long_name = dev->short_name = calloc (256, sizeof (char));
+  {
+     char *name = calloc (256, sizeof (char));
 
-  if (ret != CL_INVALID_DEVICE)
-    cuDeviceGetName (dev->long_name, 256, data->device);
-  else
-    snprintf (dev->long_name, 255, "Unavailable CUDA device #%d", j);
+     if (ret != CL_INVALID_DEVICE)
+       cuDeviceGetName (name, 256, data->device);
+     else
+       snprintf (name, 255, "Unavailable CUDA device #%d", j);
+     dev->long_name = dev->short_name = name;
+  }
 
   SETUP_DEVICE_CL_VERSION (CUDA_DEVICE_CL_VERSION_MAJOR,
                            CUDA_DEVICE_CL_VERSION_MINOR);
@@ -325,6 +334,16 @@ pocl_cuda_init (unsigned j, cl_device_id dev, const char *parameters)
       dev->max_clock_frequency /= 1000;
       GET_CU_PROP (TEXTURE_ALIGNMENT, dev->mem_base_addr_align);
       GET_CU_PROP (INTEGRATED, dev->host_unified_memory);
+#if CUDA_VERSION >= 11010
+      GET_CU_PROP (READ_ONLY_HOST_REGISTER_SUPPORTED, data->supports_cu_mem_host_register);
+#elif defined(__aarch64__) || defined(__arm__)
+      // For cuda < 11.1, we don't know if the device supports cuMemHostRegister
+      // or not. Let's assume that it doesn't in ARM devices.
+      // This gives a false negative for Jetson Xavier, but it is the best we could do.
+      data->supports_cu_mem_host_register = 0;
+#else
+      data->supports_cu_mem_host_register = 1;
+#endif
     }
   if (CUDA_CHECK_ERROR (result, "cuDeviceGetAttribute"))
     ret = CL_INVALID_DEVICE;
@@ -409,6 +428,11 @@ pocl_cuda_init (unsigned j, cl_device_id dev, const char *parameters)
   dev->max_mem_alloc_size = max (memtotal / 4, 128 * 1024 * 1024);
   dev->global_mem_size = memtotal;
 
+  // All devices starting from Compute Capability 2.0 have this limit;
+  // See e.g.
+  // https://forums.developer.nvidia.com/t/max-size-of-cuda-arguments/50218
+  dev->max_parameter_size = 4352;
+
   dev->data = data;
 
   POCL_INIT_LOCK (data->compile_lock);
@@ -435,9 +459,9 @@ pocl_cuda_init_queue (cl_device_id device, cl_command_queue queue)
 
   if (queue_data->use_threads)
     {
-      pthread_mutex_init (&queue_data->lock, NULL);
-      pthread_cond_init (&queue_data->pending_cond, NULL);
-      pthread_cond_init (&queue_data->running_cond, NULL);
+      PTHREAD_CHECK (pthread_mutex_init (&queue_data->lock, NULL));
+      PTHREAD_CHECK (pthread_cond_init (&queue_data->pending_cond, NULL));
+      PTHREAD_CHECK (pthread_cond_init (&queue_data->running_cond, NULL));
       int err = pthread_create (&queue_data->submit_thread, NULL,
                                 pocl_cuda_submit_thread, queue_data);
       if (err)
@@ -472,13 +496,13 @@ pocl_cuda_free_queue (cl_device_id device, cl_command_queue queue)
   /* Kill queue threads */
   if (queue_data->use_threads)
     {
-      pthread_mutex_lock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_lock (&queue_data->lock));
       queue_data->queue = NULL;
-      pthread_cond_signal (&queue_data->pending_cond);
-      pthread_cond_signal (&queue_data->running_cond);
-      pthread_mutex_unlock (&queue_data->lock);
-      pthread_join (queue_data->submit_thread, NULL);
-      pthread_join (queue_data->finalize_thread, NULL);
+      PTHREAD_CHECK (pthread_cond_signal (&queue_data->pending_cond));
+      PTHREAD_CHECK (pthread_cond_signal (&queue_data->running_cond));
+      PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
+      PTHREAD_CHECK (pthread_join (queue_data->submit_thread, NULL));
+      PTHREAD_CHECK (pthread_join (queue_data->finalize_thread, NULL));
     }
   return CL_SUCCESS;
 }
@@ -525,13 +549,18 @@ pocl_cuda_uninit (unsigned j, cl_device_id device)
 {
   pocl_cuda_device_data_t *data = device->data;
 
-  if (device->available)
+  if (device->available) {
+      cuEventDestroy (data->epoch_event);
       cuCtxDestroy (data->context);
+  }
 
   POCL_MEM_FREE (data);
   device->data = NULL;
 
-  POCL_MEM_FREE (device->long_name);
+  char *name = (char*)device->long_name;
+  POCL_MEM_FREE (name);
+  device->long_name = device->short_name = NULL;
+
   return CL_SUCCESS;
 }
 
@@ -551,30 +580,33 @@ pocl_cuda_alloc_mem_obj (cl_device_id device, cl_mem mem, void *host_ptr)
 
   if (flags & CL_MEM_USE_HOST_PTR)
     {
-#if defined __arm__
-          /* cuMemHostRegister is not supported on ARM.
+      if (!((pocl_cuda_device_data_t *)device->data)->supports_cu_mem_host_register)
+        {
+          /* cuMemHostRegister is not supported on some ARM devices like the Nano, but supported on Xavier.
            * Allocate device memory and perform explicit copies
            * before and after running a kernel */
-          result = cuMemAlloc ((CUdeviceptr *)&b, mem_obj->size);
+          result = cuMemAlloc ((CUdeviceptr *)&b, mem->size);
           CUDA_CHECK (result, "cuMemAlloc");
-#else
-      POCL_RETURN_ERROR_ON ((pocl_alloc_or_retain_mem_host_ptr (mem) != 0),
-                            CL_OUT_OF_HOST_MEMORY,
-                            "Cannot allocate backing memory!\n");
+        }
+      else
+        {
+          POCL_RETURN_ERROR_ON ((pocl_alloc_or_retain_mem_host_ptr (mem) != 0),
+                                CL_OUT_OF_HOST_MEMORY,
+                                "Cannot allocate backing memory!\n");
 
-      result = cuMemHostRegister (mem->mem_host_ptr, mem->size,
-                                  CU_MEMHOSTREGISTER_DEVICEMAP);
-      if (result != CUDA_SUCCESS
-          && result != CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED)
-        CUDA_CHECK (result, "cuMemHostRegister");
-      result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b, mem->mem_host_ptr,
-                                          0);
-      CUDA_CHECK (result, "cuMemHostGetDevicePointer");
+          result = cuMemHostRegister (mem->mem_host_ptr, mem->size,
+                                      CU_MEMHOSTREGISTER_DEVICEMAP);
+          if (result != CUDA_SUCCESS
+              && result != CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED)
+            CUDA_CHECK (result, "cuMemHostRegister");
+          result = cuMemHostGetDevicePointer ((CUdeviceptr *)&b, mem->mem_host_ptr,
+                                              0);
+          CUDA_CHECK (result, "cuMemHostGetDevicePointer");
 
-      /* TODO can we assume cuMemHostRegister copies
-       * the content of host memory to the device ? for now, lets not */
-      p->version = 0;
-#endif
+          /* TODO can we assume cuMemHostRegister copies
+           * the content of host memory to the device ? for now, lets not */
+          p->version = 0;
+        }
     }
   /* preallocate host visible memory */
   else if ((flags & CL_MEM_ALLOC_HOST_PTR) && (mem->mem_host_ptr == NULL))
@@ -627,12 +659,15 @@ pocl_cuda_free (cl_device_id device, cl_mem mem_obj)
 
   if (mem_obj->flags & CL_MEM_USE_HOST_PTR)
     {
-#if defined __arm__
-      cuMemFree ((CUdeviceptr)p->mem_ptr);
-#else
-      assert (p->extra_ptr == NULL);
-      cuMemHostUnregister (mem_obj->mem_host_ptr);
-#endif
+      if (((pocl_cuda_device_data_t *)device->data)->supports_cu_mem_host_register)
+        {
+          assert (p->extra_ptr == NULL);
+          cuMemHostUnregister (mem_obj->mem_host_ptr);
+        }
+      else
+        {
+          cuMemFree ((CUdeviceptr)p->mem_ptr);
+        }
     }
   else if (p->extra_ptr)
     {
@@ -942,8 +977,47 @@ load_or_generate_kernel (cl_kernel kernel, cl_device_id device,
   /* Load PTX module */
   /* TODO: When can we unload the module? */
   CUmodule module;
-  result = cuModuleLoad (&module, ptx_filename);
-  CUDA_CHECK (result, "cuModuleLoad");
+
+  if (!(pocl_debug_messages_filter & POCL_DEBUG_FLAG_CUDA))
+    {
+      result = cuModuleLoad (&module, ptx_filename);
+      CUDA_CHECK (result, "cuModuleLoad");
+    }
+  else
+    {
+      struct stat st;
+      stat (ptx_filename, &st);
+
+      const int fd = open (ptx_filename, O_RDONLY);
+      assert (fd != -1);
+
+      char *buffer = (char *)malloc (st.st_size + 1);
+      const ssize_t r = read (fd, buffer, st.st_size);
+      assert (r == st.st_size);
+
+      buffer[r] = '\0';
+      close (fd);
+
+      unsigned int log_size = 1 << 12;
+      char *log = (char *)malloc (log_size);
+
+      CUjit_option opt[]
+          = { CU_JIT_ERROR_LOG_BUFFER, CU_JIT_ERROR_LOG_BUFFER_SIZE_BYTES };
+      void *val[] = { log, (void *)log_size };
+      result = cuModuleLoadDataEx (&module, buffer,
+                                   sizeof (opt) / sizeof (opt[0]), opt, val);
+
+      unsigned int out_size = (unsigned int)val[1];
+
+      if (out_size > 0 || result != CUDA_SUCCESS)
+        POCL_MSG_PRINT_CUDA ("cuModuleLoadDataEx(%s) log: %s\n", ptx_filename,
+                             log);
+
+      CUDA_CHECK (result, "cuModuleLoadDataEx");
+
+      free (log);
+      free (buffer);
+    }
 
   /* Get kernel function */
   CUfunction function;
@@ -979,7 +1053,7 @@ void
 pocl_cuda_compile_kernel (_cl_command_node *cmd, cl_kernel kernel,
                           cl_device_id device, int specialize)
 {
-  load_or_generate_kernel (kernel, device, 0, cmd->device_i, cmd,
+  load_or_generate_kernel (kernel, device, 0, cmd->program_device_i, cmd,
                            specialize);
 }
 
@@ -993,13 +1067,16 @@ pocl_cuda_submit_kernel (CUstream stream, _cl_command_node *cmd,
   struct pocl_context pc = run.pc;
   pocl_kernel_metadata_t *meta = kernel->meta;
 
+  if (pc.num_groups[0] == 0 || pc.num_groups[1] == 0 || pc.num_groups[2] == 0)
+    return;
+
   /* Check if we need to handle global work offsets */
   int has_offsets =
     (pc.global_offset[0] || pc.global_offset[1] || pc.global_offset[2]);
 
   /* Get kernel function */
   pocl_cuda_kernel_data_t *kdata = load_or_generate_kernel (
-      kernel, device, has_offsets, cmd->device_i, cmd, 1);
+      kernel, device, has_offsets, cmd->program_device_i, cmd, 1);
   CUmodule module = has_offsets ? kdata->module_offsets : kdata->module;
   CUfunction function = has_offsets ? kdata->kernel_offsets : kdata->kernel;
 
@@ -1082,16 +1159,15 @@ pocl_cuda_submit_kernel (CUstream stream, _cl_command_node *cmd,
                     params[i] = &mem->device_ptrs[device->global_mem_id].mem_ptr
                                 + arguments[i].offset;
 
-#if defined __arm__
                     /* On ARM with USE_HOST_PTR, perform explicit copy to
                      * device */
-                    if (mem->flags & CL_MEM_USE_HOST_PTR)
+                    if ((mem->flags & CL_MEM_USE_HOST_PTR) &&
+                        !((pocl_cuda_device_data_t *)device->data)->supports_cu_mem_host_register)
                       {
                         cuMemcpyHtoD (*(CUdeviceptr *)(params[i]),
                                       mem->mem_host_ptr, mem->size);
                         cuStreamSynchronize (0);
                       }
-#endif
                   }
                 else
                   {
@@ -1405,14 +1481,14 @@ pocl_cuda_submit (_cl_command_node *node, cl_command_queue cq)
   if (((pocl_cuda_queue_data_t *)cq->data)->use_threads)
     {
 
-      pthread_cond_init (&p->event_cond, NULL);
+      PTHREAD_CHECK (pthread_cond_init (&p->event_cond, NULL));
       /* Add command to work queue */
       POCL_UNLOCK_OBJ (node->event);
       pocl_cuda_queue_data_t *queue_data = (pocl_cuda_queue_data_t *)cq->data;
-      pthread_mutex_lock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_lock (&queue_data->lock));
       DL_APPEND (queue_data->pending_queue, node);
-      pthread_cond_signal (&queue_data->pending_cond);
-      pthread_mutex_unlock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_cond_signal (&queue_data->pending_cond));
+      PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
     }
   else
     {
@@ -1466,50 +1542,39 @@ pocl_cuda_finalize_command (cl_device_id device, cl_event event)
   result = cuEventSynchronize (event_data->end);
   CUDA_CHECK (result, "cuEventSynchronize");
 
-  /* Clean up mapped memory allocations */
-  if (event->command_type == CL_COMMAND_UNMAP_MEM_OBJECT)
-    {
-      pocl_unmap_command_finished2 (event, &event->command->command);
-    }
-
   if (event->command_type == CL_COMMAND_NDRANGE_KERNEL
       || event->command_type == CL_COMMAND_TASK)
     {
-#if defined __arm__
-      /* On ARM with USE_HOST_PTR, perform explict copies back from device */
-      cl_kernel kernel = event->command.run.kernel;
-      pocl_argument *arguments = event->command.run.arguments;
-      unsigned i;
-      for (i = 0; i < meta->num_args; i++)
+      if (!((pocl_cuda_device_data_t *)device->data)->supports_cu_mem_host_register)
         {
-          if (meta->arg_info[i].type == POCL_ARG_TYPE_POINTER)
+          /* On ARM with USE_HOST_PTR, perform explict copies back from device */
+          cl_kernel kernel = event->command->command.run.kernel;
+          pocl_argument *arguments = event->command->command.run.arguments;
+          unsigned i;
+          pocl_kernel_metadata_t *meta = kernel->meta;
+          for (i = 0; i < meta->num_args; i++)
             {
-              if (!ARG_IS_LOCAL (meta->arg_info[i]) && arguments[i].value)
+              if (meta->arg_info[i].type == POCL_ARG_TYPE_POINTER)
                 {
-                  cl_mem mem = *(void **)arguments[i].value;
-                  if (mem->flags & CL_MEM_USE_HOST_PTR)
+                  if (!ARG_IS_LOCAL (meta->arg_info[i]) && arguments[i].value)
                     {
-                      CUdeviceptr ptr
-                          = (CUdeviceptr)mem->device_ptrs[device->global_mem_id]
-                                .mem_ptr;
-                      cuMemcpyDtoH (mem->mem_host_ptr, ptr, mem->size);
-                      cuStreamSynchronize (0);
+                      cl_mem mem = *(void **)arguments[i].value;
+                      if (mem->flags & CL_MEM_USE_HOST_PTR)
+                        {
+                          CUdeviceptr ptr
+                              = (CUdeviceptr)mem->device_ptrs[device->global_mem_id]
+                                    .mem_ptr;
+                          cuMemcpyDtoH (mem->mem_host_ptr, ptr, mem->size);
+                          cuStreamSynchronize (0);
+                        }
                     }
                 }
             }
         }
-#endif
-
-      pocl_ndrange_node_cleanup (event->command);
-    }
-  else
-    {
-      pocl_mem_manager_free_command (event->command);
     }
 
   /* Handle failed events */
 
-
   pocl_update_event_running (event);
   if (event->status < 0)
     pocl_update_event_failed (event);
@@ -1544,14 +1609,14 @@ pocl_cuda_update_event (cl_device_id device, cl_event event)
           &diff, ((pocl_cuda_device_data_t *)device->data)->epoch_event,
           event_data->start);
       CUDA_CHECK (result, "cuEventElapsedTime");
-      event->time_start = (cl_ulong) (epoch + diff * 1e6);
+      event->time_start = epoch + (cl_ulong)(diff * 1e6);
       event->time_start = max (event->time_start, epoch + 1);
 
       result = cuEventElapsedTime (
           &diff, ((pocl_cuda_device_data_t *)device->data)->epoch_event,
           event_data->end);
       CUDA_CHECK (result, "cuEventElapsedTime");
-      event->time_end = (cl_ulong) (epoch + diff * 1e6);
+      event->time_end = epoch + (cl_ulong)(diff * 1e6);
       event->time_end = max (event->time_end, event->time_start + 1);
     }
 }
@@ -1572,7 +1637,7 @@ pocl_cuda_notify_event_finished (cl_event event)
   pocl_cuda_event_data_t *e_d = (pocl_cuda_event_data_t *)event->data;
 
   if (((pocl_cuda_queue_data_t *)event->queue->data)->use_threads)
-    pthread_cond_broadcast (&e_d->event_cond);
+    PTHREAD_CHECK (pthread_cond_broadcast (&e_d->event_cond));
 }
 
 void
@@ -1586,7 +1651,8 @@ pocl_cuda_wait_event (cl_device_id device, cl_event event)
       POCL_LOCK_OBJ (event);
       while (event->status > CL_COMPLETE)
         {
-          pthread_cond_wait (&e_d->event_cond, &event->pocl_lock);
+          PTHREAD_CHECK (
+              pthread_cond_wait (&e_d->event_cond, &event->pocl_lock));
         }
       POCL_UNLOCK_OBJ (event);
     }
@@ -1604,7 +1670,7 @@ pocl_cuda_free_event_data (cl_event event)
     {
       pocl_cuda_event_data_t *event_data
           = (pocl_cuda_event_data_t *)event->data;
-      pthread_cond_destroy (&event_data->event_cond);
+      PTHREAD_CHECK (pthread_cond_destroy (&event_data->event_cond));
       if (event->queue->properties & CL_QUEUE_PROFILING_ENABLE)
         cuEventDestroy (event_data->start);
       cuEventDestroy (event_data->end);
@@ -1653,22 +1719,23 @@ pocl_cuda_submit_thread (void *data)
     {
       /* Attempt to get next command from work queue */
       _cl_command_node *node = NULL;
-      pthread_mutex_lock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_lock (&queue_data->lock));
       if (!queue_data->queue)
         {
-          pthread_mutex_unlock (&queue_data->lock);
+          PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
           break;
         }
       if (!queue_data->pending_queue)
         {
-          pthread_cond_wait (&queue_data->pending_cond, &queue_data->lock);
+          PTHREAD_CHECK (pthread_cond_wait (&queue_data->pending_cond,
+                                            &queue_data->lock));
         }
       if (queue_data->pending_queue)
         {
           node = queue_data->pending_queue;
           DL_DELETE (queue_data->pending_queue, node);
         }
-      pthread_mutex_unlock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
 
       /* Submit command, if we found one */
       if (node)
@@ -1676,10 +1743,10 @@ pocl_cuda_submit_thread (void *data)
           pocl_cuda_submit_node (node, queue_data->queue, 0);
 
           /* Add command to running queue */
-          pthread_mutex_lock (&queue_data->lock);
+          PTHREAD_CHECK (pthread_mutex_lock (&queue_data->lock));
           DL_APPEND (queue_data->running_queue, node);
-          pthread_cond_signal (&queue_data->running_cond);
-          pthread_mutex_unlock (&queue_data->lock);
+          PTHREAD_CHECK (pthread_cond_signal (&queue_data->running_cond));
+          PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
         }
     }
 
@@ -1703,22 +1770,23 @@ pocl_cuda_finalize_thread (void *data)
     {
       /* Attempt to get next node from running queue */
       _cl_command_node *node = NULL;
-      pthread_mutex_lock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_lock (&queue_data->lock));
       if (!queue_data->queue)
         {
-          pthread_mutex_unlock (&queue_data->lock);
+          PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
           break;
         }
       if (!queue_data->running_queue)
         {
-          pthread_cond_wait (&queue_data->running_cond, &queue_data->lock);
+          PTHREAD_CHECK (pthread_cond_wait (&queue_data->running_cond,
+                                            &queue_data->lock));
         }
       if (queue_data->running_queue)
         {
           node = queue_data->running_queue;
           DL_DELETE (queue_data->running_queue, node);
         }
-      pthread_mutex_unlock (&queue_data->lock);
+      PTHREAD_CHECK (pthread_mutex_unlock (&queue_data->lock));
 
       /* Wait for command to finish, if we found one */
       if (node)
