@@ -38,8 +38,12 @@
 #include "llvm/IR/Module.h"
 #include "llvm/IR/Verifier.h"
 #include "llvm/Linker/Linker.h"
-#include "llvm/Support/FileSystem.h"
+#ifndef LLVM_OLDER_THAN_14_0
+#include "llvm/MC/TargetRegistry.h"
+#else
 #include "llvm/Support/TargetRegistry.h"
+#endif
+#include "llvm/Support/FileSystem.h"
 #include "llvm/Support/TargetSelect.h"
 #ifndef LLVM_OLDER_THAN_11_0
 #include "llvm/Support/Alignment.h"
@@ -325,7 +329,12 @@ void fixPrintF(llvm::Module *Module) {
     if (!Call)
       continue;
 
+#ifndef LLVM_OLDER_THAN_8_0
+    unsigned NumArgs = Call->arg_size() - 1;
+#else
     unsigned NumArgs = Call->getNumArgOperands() - 1;
+#endif
+
     llvm::Value *Format = Call->getArgOperand(0);
 
     // Allocate array for arguments.
@@ -753,8 +762,8 @@ void fixConstantMemArgs(llvm::Module *Module, const char *KernelName) {
       llvm::ArrayType::get(llvm::Type::getInt8Ty(Module->getContext()),
                            65536 - TotalAutoConstantSize);
   llvm::GlobalVariable *ConstantMemBase = new llvm::GlobalVariable(
-      *Module, ByteArrayType, false, llvm::GlobalValue::InternalLinkage,
-      llvm::Constant::getNullValue(ByteArrayType), "_constant_memory_region_",
+      *Module, ByteArrayType, false, llvm::GlobalValue::ExternalLinkage,
+      NULL, "_constant_memory_region_",
       NULL, llvm::GlobalValue::NotThreadLocal, 4, false);
 
   convertPtrArgsToOffsets(Module, KernelName, 4, ConstantMemBase);
@@ -854,6 +863,7 @@ void mapLibDeviceCalls(llvm::Module *Module) {
 
     {"remquo", "__nv_remquo"},
     {"remquof", "__nv_remquof"},
+
     // TODO: lgamma_r
     // TODO: rootn
   };
