@@ -78,6 +78,14 @@
 #include "accel/accel.h"
 #endif
 
+#ifdef BUILD_PROXY
+#include "proxy/pocl_proxy.h"
+#endif
+
+#ifdef BUILD_VULKAN
+#include "vulkan/pocl-vulkan.h"
+#endif
+
 #define MAX_DEV_NAME_LEN 64
 
 #ifndef PATH_MAX
@@ -91,8 +99,9 @@
 #include <dlfcn.h>
 #endif
 
+
 /* the enabled devices */
-static struct _cl_device_id* pocl_devices = NULL;
+struct _cl_device_id *pocl_devices = NULL;
 unsigned int pocl_num_devices = 0;
 
 #ifdef ENABLE_LOADABLE_DRIVERS
@@ -101,6 +110,15 @@ unsigned int pocl_num_devices = 0;
 #define INIT_DEV(ARG) pocl_##ARG##_init_device_ops
 #endif
 
+const char *
+pocl_get_device_name (unsigned index)
+{
+  if (index < pocl_num_devices)
+    return pocl_devices[index].long_name;
+  else
+    return NULL;
+}
+
 /* Init function prototype */
 typedef void (*init_device_ops)(struct pocl_device_ops*);
 
@@ -124,6 +142,12 @@ static init_device_ops pocl_devices_init_ops[] = {
 #ifdef BUILD_ACCEL
   INIT_DEV (accel),
 #endif
+#ifdef BUILD_PROXY
+  INIT_DEV (proxy),
+#endif
+#ifdef BUILD_VULKAN
+  INIT_DEV (vulkan),
+#endif
 };
 
 #define POCL_NUM_DEVICE_TYPES (sizeof(pocl_devices_init_ops) / sizeof((pocl_devices_init_ops)[0]))
@@ -147,6 +171,12 @@ char pocl_device_types[POCL_NUM_DEVICE_TYPES][30] = {
 #ifdef BUILD_ACCEL
   "accel",
 #endif
+#ifdef BUILD_PROXY
+  "proxy",
+#endif
+#ifdef BUILD_VULKAN
+  "vulkan",
+#endif
 };
 
 static struct pocl_device_ops pocl_device_ops[POCL_NUM_DEVICE_TYPES];
