@@ -257,7 +257,7 @@ pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
   ops->svm_free = pocl_hsa_svm_free;
   ops->svm_alloc = pocl_hsa_svm_alloc;
   ops->svm_copy = pocl_hsa_svm_copy;
-  ops->svm_fill = pocl_basic_svm_fill;
+  ops->svm_fill = pocl_driver_svm_fill;
   ops->svm_register = pocl_hsa_svm_register;
   ops->svm_unregister = pocl_hsa_svm_unregister;
 
@@ -485,10 +485,11 @@ get_hsa_device_features(char* dev_name, struct _cl_device_id* dev)
               /* TODO: Add a CMake variable or HSA description string
                  autodetection to control these. */
               if (dev->llvm_cpu == NULL)
-                dev->llvm_cpu = get_llvm_cpu_name ();
+                dev->llvm_cpu = pocl_get_llvm_cpu_name ();
               if (dev->llvm_target_triplet == NULL)
                 dev->llvm_target_triplet = OCL_KERNEL_TARGET;
               dev->arg_buffer_launcher = CL_TRUE;
+              dev->grid_launcher = CL_TRUE;
             }
           COPY_ATTR (has_64bit_long);
           COPY_ATTR (vendor_id);
@@ -681,6 +682,7 @@ pocl_hsa_init (unsigned j, cl_device_id dev, const char *parameters)
 
   dev->spmd = CL_TRUE;
   dev->arg_buffer_launcher = CL_FALSE;
+  dev->grid_launcher = CL_FALSE;
   dev->autolocals_to_args = POCL_AUTOLOCALS_TO_ARGS_NEVER;
   dev->device_alloca_locals = CL_FALSE;
 
@@ -811,15 +813,15 @@ pocl_hsa_init (unsigned j, cl_device_id dev, const char *parameters)
                   | CL_DEVICE_SVM_FINE_GRAIN_BUFFER
                   | CL_DEVICE_SVM_ATOMICS;
   /* This is from clinfo output ran on AMD Catalyst drivers */
-  dev->max_events = 1024;
-  dev->max_queues = 1;
+  dev->pipe_support = CL_FALSE;
+  dev->max_events = 0;
+  dev->max_queues = 0;
   dev->max_pipe_args = 16;
   dev->max_pipe_active_res = 16;
   dev->max_pipe_packet_size = 1024 * 1024;
-  dev->dev_queue_pref_size = 256 * 1024;
-  dev->dev_queue_max_size = 512 * 1024;
-  dev->on_dev_queue_props
-      = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_PROFILING_ENABLE;
+  dev->dev_queue_pref_size = 0;
+  dev->dev_queue_max_size = 0;
+  dev->on_dev_queue_props = 0;
   dev->on_host_queue_props = CL_QUEUE_PROFILING_ENABLE;
 
   pocl_hsa_device_data_t *d = init_dev_data (dev, j);
@@ -879,7 +881,7 @@ pocl_hsa_alloc_mem_obj (cl_device_id device, cl_mem mem_obj, void *host_ptr)
 {
   /* if we share global memory with CPU, let the CPU driver allocate it */
   if (device->global_mem_id == 0)
-    return pocl_basic_alloc_mem_obj (device, mem_obj, host_ptr);
+    return pocl_driver_alloc_mem_obj (device, mem_obj, host_ptr);
 
   /* ... otherwise allocate it via HSA. */
   pocl_mem_identifier *p = &mem_obj->device_ptrs[device->global_mem_id];
@@ -900,7 +902,7 @@ pocl_hsa_free (cl_device_id device, cl_mem memobj)
 {
   /* if we share global memory with CPU, let the CPU driver free it */
   if (device->global_mem_id == 0)
-    return pocl_basic_free (device, memobj);
+    return pocl_driver_free (device, memobj);
 
   /* ... otherwise free it via HSA. */
   cl_mem_flags flags = memobj->flags;
@@ -1075,7 +1077,7 @@ compile_parallel_bc_to_brig (char *brigfile, _cl_command_node *cmd,
   _cl_command_run *run_cmd = &cmd->command.run;
 
   pocl_cache_work_group_function_path (parallel_bc_path,
-                                       run_cmd->kernel->program, cmd->device_i,
+                                       run_cmd->kernel->program, cmd->program_device_i,
                                        run_cmd->kernel, cmd, specialize);
 
   strcpy (brigfile, parallel_bc_path);
@@ -1295,7 +1297,7 @@ pocl_hsa_compile_kernel_hsail (_cl_command_node *cmd, cl_kernel kernel,
 
   POCL_LOCK (d->pocl_hsa_compilation_lock);
 
-  int error = pocl_llvm_generate_workgroup_function (cmd->device_i, device,
+  int error = pocl_llvm_generate_workgroup_function (cmd->program_device_i, device,
                                                      kernel, cmd, specialize);
   if (error)
     {
@@ -1680,6 +1682,15 @@ pocl_hsa_launch (pocl_hsa_device_data_t *d, cl_event event)
   _cl_command_run *run_cmd = &cmd->command.run;
   cl_kernel kernel = cmd->command.run.kernel;
   struct pocl_context *pc = &cmd->command.run.pc;
+
+  if (pc->num_groups[0] == 0 || pc->num_groups[1] == 0 || pc->num_groups[2] == 0)
+    {
+      pocl_update_event_running_unlocked (event);
+      POCL_UNLOCK_OBJ (event);
+      POCL_UPDATE_EVENT_COMPLETE (event);
+      return;
+    }
+
   hsa_kernel_dispatch_packet_t *kernel_packet;
   pocl_hsa_device_pthread_data_t* dd = &d->driver_data;
   pocl_hsa_event_data_t *event_data = (pocl_hsa_event_data_t *)event->data;
@@ -1857,9 +1868,6 @@ pocl_hsa_ndrange_event_finished (pocl_hsa_device_data_t *d, size_t i)
     }
 
   POCL_UPDATE_EVENT_COMPLETE (event);
-
-  pocl_ndrange_node_cleanup (node);
-  pocl_mem_manager_free_command (node);
 }
 
 static void
@@ -2036,7 +2044,7 @@ void
 pocl_hsa_notify_event_finished (cl_event event)
 {
   pocl_hsa_event_data_t *e_d = (pocl_hsa_event_data_t *)event->data;
-  pthread_cond_broadcast (&e_d->event_cond);
+  PTHREAD_CHECK (pthread_cond_broadcast (&e_d->event_cond));
 }
 
 void
@@ -2049,7 +2057,7 @@ pocl_hsa_update_event (cl_device_id device, cl_event event)
       pocl_hsa_event_data_t *e_d
           = (pocl_hsa_event_data_t *)malloc (sizeof (pocl_hsa_event_data_t));
       assert (e_d);
-      pthread_cond_init (&e_d->event_cond, NULL);
+      PTHREAD_CHECK (pthread_cond_init (&e_d->event_cond, NULL));
       event->data = (void *)e_d;
     }
   else
