@@ -36,7 +36,7 @@
 typedef struct pool_thread_data thread_data;
 
 /* Initializes scheduler. Must be called before any kernel enqueue */
-void pthread_scheduler_init (cl_device_id device);
+cl_int pthread_scheduler_init (cl_device_id device);
 
 void pthread_scheduler_uninit ();
 
