@@ -40,6 +40,9 @@
 #include "pocl_util.h"
 #include "common.h"
 #include "pocl_mem_management.h"
+#ifdef __APPLE__
+#include "pthread_barrier.h"
+#endif
 
 static void* pocl_pthread_driver_thread (void *p);
 
@@ -77,18 +80,23 @@ typedef struct scheduler_data_
   POCL_FAST_LOCK_T wq_lock_fast __attribute__ ((aligned (HOST_CPU_CACHELINE_SIZE)));
 
   int thread_pool_shutdown_requested;
+
+  pthread_barrier_t init_barrier
+      __attribute__ ((aligned (HOST_CPU_CACHELINE_SIZE)));
+
+  int worker_out_of_memory;
 } scheduler_data __attribute__ ((aligned (HOST_CPU_CACHELINE_SIZE)));
 
 static scheduler_data scheduler;
 
-void
+cl_int
 pthread_scheduler_init (cl_device_id device)
 {
   unsigned i;
   size_t num_worker_threads = device->max_compute_units;
   POCL_FAST_INIT (scheduler.wq_lock_fast);
 
-  pthread_cond_init (&(scheduler.wake_pool), NULL);
+  PTHREAD_CHECK (pthread_cond_init (&(scheduler.wake_pool), NULL));
 
   scheduler.thread_pool = pocl_aligned_malloc (
       HOST_CPU_CACHELINE_SIZE,
@@ -106,14 +114,28 @@ pthread_scheduler_init (cl_device_id device)
    * TODO fix this */
   scheduler.local_mem_size = device->local_mem_size + device->max_parameter_size * MAX_EXTENDED_ALIGNMENT;
 
+  PTHREAD_CHECK (pthread_barrier_init (&scheduler.init_barrier, NULL,
+                                       num_worker_threads + 1));
+  scheduler.worker_out_of_memory = 0;
+
   for (i = 0; i < num_worker_threads; ++i)
     {
       scheduler.thread_pool[i].index = i;
-      pthread_create (&scheduler.thread_pool[i].thread, NULL,
-                      pocl_pthread_driver_thread,
-                      (void*)&scheduler.thread_pool[i]);
+      PTHREAD_CHECK (pthread_create (&scheduler.thread_pool[i].thread, NULL,
+                                     pocl_pthread_driver_thread,
+                                     (void *)&scheduler.thread_pool[i]));
     }
 
+  PTHREAD_CHECK2 (PTHREAD_BARRIER_SERIAL_THREAD,
+                  pthread_barrier_wait (&scheduler.init_barrier));
+
+  if (scheduler.worker_out_of_memory)
+    {
+      pthread_scheduler_uninit ();
+      return CL_OUT_OF_HOST_MEMORY;
+    }
+
+  return CL_SUCCESS;
 }
 
 void
@@ -123,17 +145,18 @@ pthread_scheduler_uninit ()
 
   POCL_FAST_LOCK (scheduler.wq_lock_fast);
   scheduler.thread_pool_shutdown_requested = 1;
-  pthread_cond_broadcast (&scheduler.wake_pool);
+  PTHREAD_CHECK (pthread_cond_broadcast (&scheduler.wake_pool));
   POCL_FAST_UNLOCK (scheduler.wq_lock_fast);
 
   for (i = 0; i < scheduler.num_threads; ++i)
     {
-      pthread_join (scheduler.thread_pool[i].thread, NULL);
+      PTHREAD_CHECK (pthread_join (scheduler.thread_pool[i].thread, NULL));
     }
 
   pocl_aligned_free (scheduler.thread_pool);
   POCL_FAST_DESTROY (scheduler.wq_lock_fast);
-  pthread_cond_destroy (&scheduler.wake_pool);
+  PTHREAD_CHECK (pthread_cond_destroy (&scheduler.wake_pool));
+  PTHREAD_CHECK (pthread_barrier_destroy (&scheduler.init_barrier));
 
   scheduler.thread_pool_shutdown_requested = 0;
 }
@@ -144,7 +167,7 @@ void pthread_scheduler_push_command (_cl_command_node *cmd)
 {
   POCL_FAST_LOCK (scheduler.wq_lock_fast);
   DL_APPEND (scheduler.work_queue, cmd);
-  pthread_cond_broadcast (&scheduler.wake_pool);
+  PTHREAD_CHECK (pthread_cond_broadcast (&scheduler.wake_pool));
   POCL_FAST_UNLOCK (scheduler.wq_lock_fast);
 }
 
@@ -153,7 +176,7 @@ pthread_scheduler_push_kernel (kernel_run_command *run_cmd)
 {
   POCL_FAST_LOCK (scheduler.wq_lock_fast);
   DL_APPEND (scheduler.kernel_queue, run_cmd);
-  pthread_cond_broadcast (&scheduler.wake_pool);
+  PTHREAD_CHECK (pthread_cond_broadcast (&scheduler.wake_pool));
   POCL_FAST_UNLOCK (scheduler.wq_lock_fast);
 }
 
@@ -334,11 +357,8 @@ finalize_kernel_command (struct pool_thread_data *thread_data,
 
   pocl_release_dlhandle_cache (k->cmd);
 
-  pocl_ndrange_node_cleanup (k->cmd);
-
   POCL_UPDATE_EVENT_COMPLETE_MSG (k->cmd->event, "NDRange Kernel        ");
 
-  pocl_mem_manager_free_command (k->cmd);
   POCL_FAST_DESTROY (k->lock);
   free_kernel_run_command (k);
 }
@@ -350,10 +370,19 @@ pocl_pthread_prepare_kernel (void *data, _cl_command_node *cmd)
   cl_kernel kernel = cmd->command.run.kernel;
   struct pocl_context *pc = &cmd->command.run.pc;
 
-  pocl_check_kernel_dlhandle_cache (cmd, 1, 1);
-
   size_t num_groups = pc->num_groups[0] * pc->num_groups[1] * pc->num_groups[2];
 
+  if (num_groups == 0)
+    {
+      pocl_update_event_running (cmd->event);
+
+      POCL_UPDATE_EVENT_COMPLETE_MSG (cmd->event, "NDRange Kernel        ");
+
+      return;
+    }
+
+  pocl_check_kernel_dlhandle_cache (cmd, 1, 1);
+
   run_cmd = new_kernel_run_command ();
   run_cmd->data = data;
   run_cmd->kernel = kernel;
@@ -486,7 +515,8 @@ RETRY:
   /* if neither a command nor a kernel was available, sleep */
   if ((cmd == NULL) && (run_cmd == NULL) && (do_exit == 0))
     {
-      pthread_cond_wait (&scheduler.wake_pool, &scheduler.wq_lock_fast);
+      PTHREAD_CHECK (
+          pthread_cond_wait (&scheduler.wake_pool, &scheduler.wq_lock_fast));
       goto RETRY;
     }
 
@@ -509,22 +539,29 @@ pocl_pthread_driver_thread (void *p)
   td->num_threads = scheduler.num_threads;
   td->printf_buffer = pocl_aligned_malloc (MAX_EXTENDED_ALIGNMENT,
                                            scheduler.printf_buf_size);
-  assert (td->printf_buffer != NULL);
 
   assert (scheduler.local_mem_size > 0);
   td->local_mem = pocl_aligned_malloc (MAX_EXTENDED_ALIGNMENT,
                                        scheduler.local_mem_size);
-  assert (td->local_mem);
 #ifdef __linux__
   if (pocl_get_bool_option ("POCL_AFFINITY", 0))
     {
       cpu_set_t set;
       CPU_ZERO (&set);
       CPU_SET (td->index, &set);
-      pthread_setaffinity_np (td->thread, sizeof (cpu_set_t), &set);
+      PTHREAD_CHECK (
+          pthread_setaffinity_np (td->thread, sizeof (cpu_set_t), &set));
     }
 #endif
 
+  if (td->printf_buffer == NULL || td->local_mem == NULL)
+    {
+      POCL_ATOMIC_INC (scheduler.worker_out_of_memory);
+    }
+
+  PTHREAD_CHECK2 (PTHREAD_BARRIER_SERIAL_THREAD,
+                  pthread_barrier_wait (&scheduler.init_barrier));
+
   while (1)
     {
       do_exit = pthread_scheduler_get_work (td);
