@@ -67,6 +67,9 @@ static const char cl_parameters[] =
   "-cl-std=CL1.2 "
   "-cl-std=CL1.1 "
   "-cl-std=CL2.0 "
+  "-cl-std=CL2.1 "
+  "-cl-std=CL2.2 "
+  "-cl-std=CL3.0 "
   "-cl-kernel-arg-info "
   "-cl-strict-aliasing "
   "-cl-denorms-are-zero "
@@ -155,7 +158,6 @@ append_to_build_log (cl_program program, unsigned device_i, const char *format,
     }                                                                         \
   while (0)
 
-
 /* options must be non-NULL.
  * modded_options[size] + link_options are preallocated outputs
  */
@@ -185,10 +187,20 @@ process_options (const char *options, char *modded_options, char *link_options,
   assert (modded_options);
   assert (compiling || linking);
 
+  char replace_me = 0;
+
   size_t i = 1; /* terminating char */
   size_t needed = 0;
-  char *temp_options = (char*) malloc (strlen (options) + 1);
-  strcpy (temp_options, options);
+  char *temp_options = (char *) malloc (strlen (options) + 1);
+
+  memset (temp_options, 0, strlen (options) + 1);
+  strncpy (temp_options, options, strlen (options));
+
+  if (pocl_escape_quoted_whitespace (temp_options, &replace_me) == -1)
+  {
+    error = CL_INVALID_BUILD_OPTIONS;
+    goto ERROR;
+  }
 
   token = strtok_r (temp_options, " ", &saveptr);
   while (token != NULL)
@@ -365,6 +377,16 @@ process_options (const char *options, char *modded_options, char *link_options,
   i = strlen (modded_options);
   if ((i > 0) && (modded_options[i - 1] == ' '))
     modded_options[i - 1] = 0;
+
+  /* put back replaced whitespaces if needed */
+  if (replace_me != 0)
+  {
+    for (size_t x = 0; x < i; x++)
+    {
+      if (modded_options[x] == replace_me) modded_options[x] = ' ';
+    }
+  }
+
 ERROR:
   POCL_MEM_FREE (temp_options);
   return error;
@@ -611,11 +633,13 @@ compile_and_link_program(int compile_program,
   POCL_GOTO_LABEL_COND (PFN_NOTIFY, (pfn_notify == NULL && user_data != NULL),
                         CL_INVALID_VALUE);
 
-  POCL_GOTO_LABEL_ON (PFN_NOTIFY, program->kernels, CL_INVALID_OPERATION,
+  POCL_LOCK_OBJ (program);
+
+  POCL_GOTO_LABEL_ON (FINISH, program->kernels, CL_INVALID_OPERATION,
                       "Program already has kernels\n");
 
   POCL_GOTO_LABEL_ON (
-      PFN_NOTIFY,
+      FINISH,
       (program->source == NULL && program->binaries == NULL
        && program->builtin_kernel_names == NULL),
       CL_INVALID_PROGRAM,
@@ -623,13 +647,11 @@ compile_and_link_program(int compile_program,
       "need "
       "to call clCreateProgramWith{Binary|Source|BuiltinKernels} first\n");
 
-  POCL_GOTO_LABEL_ON (PFN_NOTIFY,
+  POCL_GOTO_LABEL_ON (FINISH,
                       ((program->source == NULL) && (link_program == 0)),
                       CL_INVALID_OPERATION,
                       "Cannot clCompileProgram when program has no source\n");
 
-  POCL_LOCK_OBJ (program);
-
   program->main_build_log[0] = 0;
 
   TP_BUILD_PROGRAM (program->context->id, program->id);
@@ -770,11 +792,13 @@ compile_and_link_program(int compile_program,
 
           if (error != CL_SUCCESS)
             {
+              if (program->build_log[device_i])
+                POCL_MSG_ERR ("Build log for device %s:\n%s\n",
+                              device->long_name, program->build_log[device_i]);
               APPEND_TO_BUILD_LOG_GOTO (build_error_code,
                                         "Device %s failed to build"
-                                        " the program, log: %s",
-                                        device->long_name,
-                                        program->build_log[device_i]);
+                                        " the program\n",
+                                        device->long_name);
             }
         }
       /* compile and/or link from binary */
@@ -800,11 +824,13 @@ compile_and_link_program(int compile_program,
 
           if (error != CL_SUCCESS)
             {
+              if (program->build_log[device_i])
+                POCL_MSG_ERR ("Build log for device %s:\n%s\n",
+                              device->long_name, program->build_log[device_i]);
               APPEND_TO_BUILD_LOG_GOTO (build_error_code,
                                         "Device %s failed to build"
-                                        " the program, log: %s",
-                                        device->long_name,
-                                        program->build_log[device_i]);
+                                        " the program\n",
+                                        device->long_name);
             }
         }
 
