@@ -111,7 +111,7 @@ void pocl_cache_program_bc_path(char*        program_bc_path,
 */
 void
 pocl_cache_kernel_cachedir_path (char *kernel_cachedir_path,
-                                 cl_program program, unsigned device_i,
+                                 cl_program program, unsigned program_device_i,
                                  cl_kernel kernel, const char *append_str,
                                  _cl_command_node *command, int specialized)
 {
@@ -137,7 +137,7 @@ pocl_cache_kernel_cachedir_path (char *kernel_cachedir_path,
       append_str);
   assert (bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
 
-  program_device_dir (kernel_cachedir_path, program, device_i, tempstring);
+  program_device_dir (kernel_cachedir_path, program, program_device_i, tempstring);
 }
 
 void
@@ -242,6 +242,14 @@ pocl_cache_write_spirv (char *spirv_path,
                               spirv_content, file_size, NULL);
 }
 
+int
+pocl_cache_write_generic_objfile (char *objfile_path,
+                                  const char *objfile_content,
+                                  uint64_t objfile_size)
+{
+  return pocl_write_tempfile (objfile_path, tempfile_pattern, ".binary",
+                              objfile_content, objfile_size, NULL);
+}
 
 /******************************************************************************/
 
@@ -269,27 +277,28 @@ int pocl_cache_device_cachedir_exists(cl_program   program,
 
 /******************************************************************************/
 
-int pocl_cache_write_descriptor(cl_program   program,
-                                unsigned     device_i,
-                                const char*  kernel_name,
-                                const char*  content,
-                                size_t       size) {
-    char devdir[POCL_FILENAME_LENGTH];
-    program_device_dir(devdir, program, device_i, "");
+int
+pocl_cache_write_descriptor (_cl_command_node *command, cl_kernel kernel,
+                             int specialize, const char *content, size_t size)
+{
+  char dirr[POCL_FILENAME_LENGTH];
 
-    char descriptor[POCL_FILENAME_LENGTH];
-    int bytes_written = snprintf(descriptor, POCL_FILENAME_LENGTH,
-                                 "%s/%s", devdir, kernel_name);
-    assert(bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
-    if (pocl_mkdir_p(descriptor))
-        return 1;
+  pocl_cache_kernel_cachedir_path (dirr, kernel->program, command->program_device_i,
+                                   kernel, "", command, specialize);
 
-    bytes_written = snprintf(descriptor, POCL_FILENAME_LENGTH,
-                                 "%s/%s/descriptor.so.kernel_obj.c",
-                                 devdir, kernel_name);
-    assert(bytes_written > 0 && bytes_written < POCL_FILENAME_LENGTH);
+  char descriptor[POCL_FILENAME_LENGTH];
 
-    return pocl_write_file(descriptor, content, size, 0, 1);
+  pocl_cache_kernel_cachedir_path (
+      descriptor, kernel->program, command->program_device_i, kernel,
+      "/../descriptor.so.kernel_obj.c", command, specialize);
+
+  if (pocl_exists (descriptor))
+    return 0;
+
+  if (pocl_mkdir_p (dirr))
+    return -1;
+
+  return pocl_write_file (descriptor, content, size, 0, 1);
 }
 
 /******************************************************************************/
