@@ -28,6 +28,7 @@
 #include "config.h"
 
 #include <assert.h>
+#include <errno.h>
 #include <stdio.h>
 
 #ifdef HAVE_VALGRIND
@@ -52,15 +53,18 @@ typedef pthread_cond_t pocl_cond_t;
 typedef pthread_t pocl_thread_t;
 #define POCL_LOCK_INITIALIZER PTHREAD_MUTEX_INITIALIZER
 
-#ifdef BUILD_ICD
-#  include "pocl_icd.h"
-#endif
 #include "pocl.h"
 #include "pocl_tracing.h"
 #include "pocl_debug.h"
 #include "pocl_hash.h"
 #include "pocl_runtime_config.h"
 #include "common.h"
+#ifdef BUILD_ICD
+#  include "pocl_icd.h"
+#endif
+
+#include <CL/cl_gl.h>
+#include <CL/cl_egl.h>
 
 #if __STDC_VERSION__ < 199901L
 # if __GNUC__ >= 2
@@ -104,40 +108,44 @@ typedef pthread_t pocl_thread_t;
 #define ALIGN_CACHE(x) x
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+POCL_EXPORT
+void pocl_abort_on_pthread_error (int status, unsigned line, const char *func);
+
+#ifdef __cplusplus
+}
+#endif
+
+/* Some pthread_*() calls may return '0' or a specific non-zero value on
+ * success.
+ */
+#define PTHREAD_CHECK2(_status_ok, _code)                                     \
+  do                                                                          \
+    {                                                                         \
+      int _pthread_status = (_code);                                          \
+      if (_pthread_status != 0 && _pthread_status != (_status_ok))            \
+        pocl_abort_on_pthread_error (_pthread_status, __LINE__,               \
+                                     __FUNCTION__);                           \
+    }                                                                         \
+  while (0)
+
+#define PTHREAD_CHECK(code) PTHREAD_CHECK2 (0, code)
+
 /* Generic functionality for handling different types of 
    OpenCL (host) objects. */
 
-#define POCL_LOCK(__LOCK__)                                                   \
-  do                                                                          \
-    {                                                                         \
-      int r = pthread_mutex_lock (&(__LOCK__));                               \
-      assert (r == 0);                                                        \
-    }                                                                         \
-  while (0)
+#define POCL_LOCK(__LOCK__) PTHREAD_CHECK (pthread_mutex_lock (&(__LOCK__)))
 #define POCL_UNLOCK(__LOCK__)                                                 \
-  do                                                                          \
-    {                                                                         \
-      int r = pthread_mutex_unlock (&(__LOCK__));                             \
-      assert (r == 0);                                                        \
-    }                                                                         \
-  while (0)
+  PTHREAD_CHECK (pthread_mutex_unlock (&(__LOCK__)))
 #define POCL_INIT_LOCK(__LOCK__)                                              \
-  do                                                                          \
-    {                                                                         \
-      int r = pthread_mutex_init (&(__LOCK__), NULL);                         \
-      assert (r == 0);                                                        \
-    }                                                                         \
-  while (0)
+  PTHREAD_CHECK (pthread_mutex_init (&(__LOCK__), NULL))
 /* We recycle OpenCL objects by not actually freeing them until the
-   very end. Thus, the lock should not be destoryed at the refcount 0. */
+   very end. Thus, the lock should not be destroyed at the refcount 0. */
 #define POCL_DESTROY_LOCK(__LOCK__)                                           \
-  do                                                                          \
-    {                                                                         \
-      int r = pthread_mutex_destroy (&(__LOCK__));                            \
-      assert (r == 0);                                                        \
-    }                                                                         \
-  while (0)
-
+  PTHREAD_CHECK (pthread_mutex_destroy (&(__LOCK__)))
 
 /* If available, use an Adaptive mutex for locking in the pthread driver,
    otherwise fallback to simple mutexes */
@@ -150,10 +158,10 @@ typedef pthread_t pocl_thread_t;
     do { \
       pthread_mutexattr_t attrs; \
       pthread_mutexattr_init (&attrs); \
-      int r = pthread_mutexattr_settype (&attrs, PTHREAD_MUTEX_ADAPTIVE_NP); \
-      assert (r == 0); \
-      pthread_mutex_init(&l, &attrs); \
-      pthread_mutexattr_destroy(&attrs);\
+      PTHREAD_CHECK (                                                         \
+          pthread_mutexattr_settype (&attrs, PTHREAD_MUTEX_ADAPTIVE_NP));     \
+      PTHREAD_CHECK (pthread_mutex_init (&l, &attrs));                        \
+      PTHREAD_CHECK (pthread_mutexattr_destroy (&attrs));                     \
     } while (0)
 #else
 #define POCL_FAST_INIT(l) POCL_INIT_LOCK (l)
@@ -161,17 +169,18 @@ typedef pthread_t pocl_thread_t;
 
 #define POCL_FAST_DESTROY(l) POCL_DESTROY_LOCK(l)
 
-#define POCL_INIT_COND(c) pthread_cond_init (&c, NULL)
-#define POCL_DESTROY_COND(c) pthread_cond_destroy (&c)
-#define POCL_SIGNAL_COND(c) pthread_cond_signal (&c)
-#define POCL_BROADCAST_COND(c) pthread_cond_broadcast (&c)
-#define POCL_WAIT_COND(c, m) pthread_cond_wait (&c, &m)
-#define POCL_TIMEDWAIT_COND(c, m, t) pthread_cond_timedwait (&c, &m, &t)
+#define POCL_INIT_COND(c) PTHREAD_CHECK (pthread_cond_init (&c, NULL))
+#define POCL_DESTROY_COND(c) PTHREAD_CHECK (pthread_cond_destroy (&c))
+#define POCL_SIGNAL_COND(c) PTHREAD_CHECK (pthread_cond_signal (&c))
+#define POCL_BROADCAST_COND(c) PTHREAD_CHECK (pthread_cond_broadcast (&c))
+#define POCL_WAIT_COND(c, m) PTHREAD_CHECK (pthread_cond_wait (&c, &m))
+#define POCL_TIMEDWAIT_COND(c, m, t) PTHREAD_CHECK2(ETIMEDOUT, pthread_cond_timedwait (&c, &m, &t))
 
 #define POCL_CREATE_THREAD(thr, func, arg)                                    \
-  pthread_create (&thr, NULL, func, arg)
-#define POCL_JOIN_THREAD(thr) pthread_join (thr, NULL)
-#define POCL_JOIN_THREAD2(thr, res_ptr) pthread_join (thr, res_ptr)
+  PTHREAD_CHECK (pthread_create (&thr, NULL, func, arg))
+#define POCL_JOIN_THREAD(thr) PTHREAD_CHECK (pthread_join (thr, NULL))
+#define POCL_JOIN_THREAD2(thr, res_ptr)                                       \
+  PTHREAD_CHECK (pthread_join (thr, res_ptr))
 #define POCL_EXIT_THREAD(res) pthread_exit (res)
 
 //############################################################################
@@ -314,13 +323,19 @@ extern uint64_t last_object_id;
 #define POdeclsym(name) __typeof__ (name) PO##name;
 #  define POCL_ALIAS_OPENCL_SYMBOL(name)                                \
   __typeof__(name) name __attribute__((alias ("PO" #name), visibility("default")));
-#  define POsymAlways(name) POCL_ALIAS_OPENCL_SYMBOL(name)
-#  if !defined(BUILD_ICD)
+
+#  if !defined(BUILD_ICD) && !defined(BUILD_PROXY)
 #    define POsym(name) POCL_ALIAS_OPENCL_SYMBOL(name)
 #  else
 #    define POsym(name)
 #  endif
 
+#  if !defined(BUILD_PROXY)
+#    define POsymAlways(name) POCL_ALIAS_OPENCL_SYMBOL(name)
+#  else
+#    define POsymAlways(name)
+#  endif
+
 #endif
 
 
@@ -503,7 +518,7 @@ struct pocl_device_ops {
   /* these are optional. If the driver needs to do anything to be able
    * to use host memory, it should do it (and undo it) in these callbacks.
    * Currently used by HSA.
-   * See pocl_basic_alloc and pocl_basic_free for details. */
+   * See pocl_driver_alloc_mem_obj and pocl_driver_free for details. */
   void (*svm_register) (cl_device_id dev, void *host_ptr, size_t size);
   void (*svm_unregister) (cl_device_id dev, void *host_ptr, size_t size);
 
@@ -513,6 +528,9 @@ struct pocl_device_ops {
                     const void *__restrict__ src, size_t size);
   void (*svm_fill) (cl_device_id dev, void *__restrict__ svm_ptr, size_t size,
                     void *__restrict__ pattern, size_t pattern_size);
+  void (*svm_migrate) (cl_device_id dev, size_t num_svm_pointers,
+                       void *__restrict__ svm_pointers,
+                       size_t *__restrict__ sizes);
 
   /* the following callbacks only deal with buffers (and IMAGE1D_BUFFER which
    * is backed by a buffer), not images.  */
@@ -690,6 +708,11 @@ struct pocl_device_ops {
   int (*init_queue) (cl_device_id device, cl_command_queue queue);
   int (*free_queue) (cl_device_id device, cl_command_queue queue);
 
+  /* Optional. if the driver needs to use hardware resources
+   * for contexts, it should use these callbacks */
+  int (*init_context) (cl_device_id device, cl_context context);
+  int (*free_context) (cl_device_id device, cl_context context);
+
   /* clEnqueueNDRangeKernel */
   void (*run) (void *data, _cl_command_node *cmd);
   /* for clEnqueueNativeKernel. may be NULL */
@@ -794,6 +817,11 @@ struct pocl_device_ops {
   cl_int (*get_device_info_ext) (cl_device_id dev, cl_device_info param_name,
                                  size_t param_value_size, void * param_value,
                                  size_t * param_value_size_ret);
+
+  /* optional. Return CL_SUCCESS if the device can be, or is associated with
+   * the GL context described in properties. */
+  cl_int (*get_gl_context_assoc) (cl_device_id device, cl_gl_context_info type,
+                                  const cl_context_properties *properties);
 };
 
 typedef struct pocl_global_mem_t {
@@ -840,6 +868,7 @@ struct _cl_device_id {
   size_t max_work_item_sizes[3];
   size_t max_work_group_size;
   size_t preferred_wg_size_multiple;
+  cl_bool non_uniform_work_group_support;
   cl_uint preferred_vector_width_char;
   cl_uint preferred_vector_width_short;
   cl_uint preferred_vector_width_int;
@@ -899,6 +928,8 @@ struct _cl_device_id {
   /* The device uses an HSA-like kernel ABI with a single argument buffer as
      an input. */
   cl_bool arg_buffer_launcher;
+  /* The device uses a GRID launcher */
+  cl_bool grid_launcher;
   /* The Workgroup pass creates launcher functions and replaces work-item
      placeholder global variables (e.g. _local_size_, _global_offset_ etc) with
      loads from the context struct passed as a kernel argument. This flag
@@ -913,16 +944,22 @@ struct _cl_device_id {
   size_t num_partition_types;
   cl_device_partition_property *partition_type;
   size_t printf_buffer_size;
-  char *short_name;
-  char *long_name;
+  const char *short_name;
+  const char *long_name;
 
   const char *vendor;
   const char *driver_version;
   const char *profile;
-  const char *version;
   const char *extensions;
-  const char *cl_version_std;  // "CL2.0"
-  cl_ulong cl_version_int;     // 200
+
+  /* these are Device versions, not OpenCL C versions */
+  const char *version;
+  unsigned version_as_int;  /* e.g. 200 */
+  cl_version version_as_cl; /* cl_version format */
+
+  /* highest OpenCL C version supported by the compiler */
+  const char *opencl_c_version_as_opt;
+  cl_version opencl_c_version_as_cl;
 
   void *data;
   const char* llvm_target_triplet; /* the llvm target triplet to use */
@@ -941,10 +978,8 @@ struct _cl_device_id {
    * the host's system time (pocl_gettimemono_ns). */
   int has_own_timer;
 
-  /* If the driver wants SPIR-V input directly, without translation to
-   * LLVM IR with "spir" triple, set this to 1,
-   * and make sure device->ops->supports_binary returns 1 for SPIR-V */
-  int consumes_il_directly;
+  /* whether this device supports OpenGL / EGL interop */
+  int has_gl_interop;
 
   /* Convert automatic local variables to kernel arguments? */
   pocl_autolocals_to_args_strategy autolocals_to_args;
@@ -1008,7 +1043,9 @@ struct _cl_device_id {
   cl_command_queue_properties on_dev_queue_props;
   cl_command_queue_properties on_host_queue_props;
   /* OpenCL 2.1 */
-  char *spirv_version;
+
+  cl_uint max_num_sub_groups;
+  cl_bool sub_group_independent_forward_progress;
 
   /* image formats supported by the device, per image type */
   const cl_image_format *image_formats[NUM_OPENCL_IMAGE_TYPES];
@@ -1017,9 +1054,45 @@ struct _cl_device_id {
   /* Device operations, shared among devices of the same type */
   struct pocl_device_ops *ops;
 
+  /* cl_khr_spir / CL_DEVICE_SPIR_VERSIONS, only includes SPIR not SPIR-V */
+  const char *supported_spir_versions;
+  /* cl_khr_il_program / CL_DEVICE_IL_VERSION, this only includes SPIR-V */
+  const char *supported_spir_v_versions;
+
   /* OpenCL 3.0 properties */
+
+  /* list of compiler features, e.g. __opencl_c_fp64 */
+  const char *features;
+
   cl_device_atomic_capabilities atomic_memory_capabilities;
   cl_device_atomic_capabilities atomic_fence_capabilities;
+
+  const char *version_of_latest_passed_cts;
+
+  cl_bool pipe_support;
+
+  /* extensions as listed in device->extensions,
+   * but with version */
+  size_t num_extensions_with_version;
+  const cl_name_version *extensions_with_version;
+
+  /* ILs and their versions supported by
+   * the compiler of the device */
+  size_t num_ils_with_version;
+  const cl_name_version *ils_with_version;
+
+  /* list of builtin kernels as in device->builtin_kernel_list,
+   * but with their versions */
+  size_t num_builtin_kernels_with_version;
+  const cl_name_version *builtin_kernels_with_version;
+
+  /* OpenCL C language versions supported by the device compiler */
+  size_t num_opencl_c_with_version;
+  const cl_name_version *opencl_c_with_version;
+
+  /* OpenCL C features supported by the device compiler */
+  size_t num_opencl_features_with_version;
+  const cl_name_version *opencl_features_with_version;
 };
 
 #define DEVICE_SVM_FINEGR(dev) (dev->svm_caps & (CL_DEVICE_SVM_FINE_GRAIN_BUFFER \
@@ -1037,12 +1110,29 @@ struct _cl_platform_id {
   POCL_ICD_OBJECT_PLATFORM_ID
 }; 
 
+typedef struct _context_destructor_callback context_destructor_callback_t;
+struct _context_destructor_callback
+{
+  void (CL_CALLBACK *pfn_notify) (cl_context, void *);
+  void *user_data;
+  context_destructor_callback_t *next;
+};
+
+typedef struct _pocl_svm_ptr pocl_svm_ptr;
+struct _pocl_svm_ptr
+{
+  void *svm_ptr;
+  size_t size;
+  struct _pocl_svm_ptr *prev, *next;
+};
+
 struct _cl_context {
   POCL_ICD_OBJECT
   POCL_OBJECT;
   /* queries */
   cl_device_id *devices;
   cl_context_properties *properties;
+  cl_bool gl_interop;
   /* implementation */
   unsigned num_devices;
   unsigned num_properties;
@@ -1082,9 +1172,19 @@ struct _cl_context {
    */
   size_t min_buffer_alignment;
 
+  /* list of destructor callbacks */
+  context_destructor_callback_t *destructor_callbacks;
+
+  /* list of SVM buffers */
+  pocl_svm_ptr *svm_ptrs;
+
 #ifdef ENABLE_LLVM
   void *llvm_context_data;
 #endif
+
+#ifdef BUILD_PROXY
+  cl_context proxied_context;
+#endif
 };
 
 typedef struct _pocl_data_sync_item pocl_data_sync_item;
@@ -1107,6 +1207,9 @@ struct _cl_command_queue {
   unsigned long command_count; /* counter for unfinished command enqueued */
   pocl_data_sync_item last_event;
 
+  cl_queue_properties queue_properties[10];
+  unsigned num_queue_properties;
+
   /* device specific data */
   void *data;
 };
@@ -1139,9 +1242,12 @@ struct _cl_command_queue {
 
 #define DEVICE_IMAGE_SIZE_SUPPORT 1
 #define DEVICE_IMAGE_FORMAT_SUPPORT 2
+#define DEVICE_IMAGE_INTEROP_SUPPORT 4
 
 #define DEVICE_DOESNT_SUPPORT_IMAGE(mem, dev_i)                               \
-  (mem->device_supports_this_image[dev_i] == 0)
+  (mem->device_supports_this_image[dev_i]                                     \
+   != (DEVICE_IMAGE_SIZE_SUPPORT | DEVICE_IMAGE_FORMAT_SUPPORT                \
+       | DEVICE_IMAGE_INTEROP_SUPPORT))
 
 #define POCL_ON_UNSUPPORTED_IMAGE(mem, dev, operation)                        \
   do                                                                          \
@@ -1151,22 +1257,25 @@ struct _cl_command_queue {
         if (mem->context->devices[dev_i] == dev)                              \
           break;                                                              \
       assert (dev_i < mem->context->num_devices);                             \
-      operation (                                                  \
-          (mem->context->devices[dev_i]->image_support == CL_FALSE),          \
-          CL_INVALID_OPERATION, "Device %s does not support images\n",        \
-          mem->context->devices[dev_i]->long_name);                           \
-      operation (                                                  \
+      operation ((mem->context->devices[dev_i]->image_support == CL_FALSE),   \
+                 CL_INVALID_OPERATION, "Device %s does not support images\n", \
+                 mem->context->devices[dev_i]->long_name);                    \
+      operation (((mem->device_supports_this_image[dev_i]                     \
+                   & DEVICE_IMAGE_FORMAT_SUPPORT)                             \
+                  == 0),                                                      \
+                 CL_IMAGE_FORMAT_NOT_SUPPORTED,                               \
+                 "The image type is not supported by this device\n");         \
+      operation (((mem->device_supports_this_image[dev_i]                     \
+                   & DEVICE_IMAGE_SIZE_SUPPORT)                               \
+                  == 0),                                                      \
+                 CL_INVALID_IMAGE_SIZE,                                       \
+                 "The image size is not supported by this device\n");         \
+      operation (                                                             \
           ((mem->device_supports_this_image[dev_i]                            \
-            & DEVICE_IMAGE_FORMAT_SUPPORT)                                    \
+            & DEVICE_IMAGE_INTEROP_SUPPORT)                                   \
            == 0),                                                             \
-          CL_IMAGE_FORMAT_NOT_SUPPORTED,                                      \
-          "The image type is not supported by this device\n");                \
-      operation (                                                  \
-          ((mem->device_supports_this_image[dev_i]                            \
-            & DEVICE_IMAGE_SIZE_SUPPORT)                                      \
-           == 0),                                                             \
-          CL_INVALID_IMAGE_SIZE,                                              \
-          "The image size is not supported by this device\n");                \
+          CL_INVALID_GL_OBJECT,                                               \
+          "OpenGL/EGL/other interop is not supported by this device\n");      \
     }                                                                         \
   while (0)
 
@@ -1187,6 +1296,9 @@ struct _cl_mem {
   cl_mem_object_type type;
   cl_mem_flags flags;
 
+  cl_mem_properties properties[5];
+  unsigned num_properties;
+
   size_t size;
   size_t origin; /* for sub-buffers */
 
@@ -1237,6 +1349,13 @@ struct _cl_mem {
   cl_mem size_buffer;
   cl_mem content_buffer;
 
+  /* OpenGL data */
+  cl_GLenum               target;
+  cl_GLint                miplevel;
+  cl_GLuint               texture;
+  CLeglDisplayKHR egl_display;
+  CLeglImageKHR egl_image;
+
   /* for images, a flag for each device in context,
    * whether that device supports this */
   int *device_supports_this_image;
@@ -1248,6 +1367,7 @@ struct _cl_mem {
 
   /* Image flags */
   cl_bool                 is_image;
+  cl_bool                 is_gl_texture;
   cl_channel_order        image_channel_order;
   cl_channel_type         image_channel_data_type;
   size_t                  image_width;
@@ -1261,14 +1381,12 @@ struct _cl_mem {
   cl_uint                 num_mip_levels;
   cl_uint                 num_samples;
   cl_mem                  buffer;
+  cl_uint                 is_gl_acquired;
 
   /* pipe flags */
   cl_bool                 is_pipe;
   size_t                  pipe_packet_size;
   size_t                  pipe_max_packets;
-
-  /* list of SVM buffers */
-  struct _cl_mem *prev, *next;
 };
 
 typedef uint8_t SHA1_digest_t[SHA1_DIGEST_SIZE * 2 + 1];
@@ -1283,6 +1401,8 @@ typedef struct pocl_kernel_metadata_s
   struct pocl_argument_info *arg_info;
   cl_bitfield has_arg_metadata;
   size_t reqd_wg_size[OPENCL_MAX_DIMENSION];
+  size_t wg_size_hint[OPENCL_MAX_DIMENSION];
+  char vectypehint[16];
 
   /* if we know the size of _every_ kernel argument, we store
    * the total size here. see struct _cl_kernel on why */
@@ -1299,6 +1419,8 @@ typedef struct pocl_kernel_metadata_s
   void **data;
 } pocl_kernel_metadata_t;
 
+#define MAIN_PROGRAM_LOG_SIZE 6400
+
 struct _cl_program {
   POCL_ICD_OBJECT
   POCL_OBJECT;
@@ -1351,7 +1473,7 @@ struct _cl_program {
   /* Per-device build logs, for the case when we don't yet have the program's cachedir */
   char** build_log;
   /* Per-program build log, for the case when we aren't yet building for devices */
-  char main_build_log[640];
+  char main_build_log[MAIN_PROGRAM_LOG_SIZE];
   /* Use to store build status */
   cl_build_status build_status;
   /* Use to store binary type */
@@ -1417,14 +1539,23 @@ struct event_node
   event_node *next;
 };
 
+#define MAX_EVENT_DEPS 60
+
 /* Optional metadata for events for improved profile data readability etc. */
 typedef struct _pocl_event_md
 {
   /* The kernel executed by the NDRange command associated with the event,
      if any. */
   cl_kernel kernel;
+
+  size_t num_deps;
+  // event IDs on which this event depends
+  uint64_t dep_ids[MAX_EVENT_DEPS];
+  // the finish time of those ^^^ event IDs
+  cl_ulong dep_ts[MAX_EVENT_DEPS];
 } pocl_event_md;
 
+
 typedef struct _cl_event _cl_event;
 struct _cl_event {
   POCL_ICD_OBJECT
@@ -1465,6 +1596,7 @@ struct _cl_event {
    * lowered and memory freed if it's 0 */
   short release_mem_host_ptr_after;
 
+
   _cl_event *next;
   _cl_event *prev;
 };
@@ -1482,6 +1614,8 @@ struct _cl_sampler {
   cl_bool             normalized_coords;
   cl_addressing_mode  addressing_mode;
   cl_filter_mode      filter_mode;
+  cl_sampler_properties properties[10];
+  cl_uint             num_properties;
   void**              device_data;
 };
 
