@@ -88,6 +88,7 @@ POdeclsym(clGetImageInfo)
 POdeclsym(clGetKernelInfo)
 POdeclsym(clGetKernelArgInfo)
 POdeclsym(clGetKernelWorkGroupInfo)
+POdeclsym(clGetKernelSubGroupInfo)
 POdeclsym(clGetMemObjectInfo)
 POdeclsym(clGetPlatformIDs)
 POdeclsym(clGetPlatformInfo)
@@ -98,7 +99,9 @@ POdeclsym(clGetSupportedImageFormats)
 POdeclsymICD(clIcdGetPlatformIDsKHR)
 POdeclsym(clReleaseCommandQueue)
 POdeclsym(clReleaseContext)
+POCL_EXPORT
 POdeclsym(clReleaseDevice)
+POCL_EXPORT
 POdeclsym(clReleaseEvent)
 POdeclsym(clReleaseKernel)
 POdeclsym(clReleaseMemObject)
@@ -107,6 +110,7 @@ POdeclsym(clReleaseSampler)
 POdeclsym(clRetainCommandQueue)
 POdeclsym(clRetainContext)
 POdeclsym(clRetainDevice)
+POCL_EXPORT
 POdeclsym(clRetainEvent)
 POdeclsym(clRetainKernel)
 POdeclsym(clRetainMemObject)
@@ -140,5 +144,18 @@ POdeclsym(clEnqueueAcquireGLObjects)
 POdeclsym(clEnqueueReleaseGLObjects)
 POdeclsym(clGetGLContextInfoKHR)
 POdeclsym(clSetContentSizeBufferPoCL)
+POdeclsym(clCreatePipe)
+POdeclsym(clGetPipeInfo)
+POdeclsym(clSetDefaultDeviceCommandQueue)
+POdeclsym(clGetDeviceAndHostTimer)
+POdeclsym(clGetHostTimer)
+POdeclsym(clSetProgramReleaseCallback)
+POdeclsym(clSetContextDestructorCallback)
+POdeclsym(clSetProgramSpecializationConstant)
+POdeclsym(clCreateSamplerWithProperties)
+POdeclsym(clCreateBufferWithProperties)
+POdeclsym(clCreateImageWithProperties)
+POdeclsym(clCloneKernel)
+POdeclsym(clEnqueueSVMMigrateMem)
 
 #endif
