@@ -36,7 +36,7 @@ extern "C" {
 
   /* Returns the cpu name as reported by LLVM. */
   POCL_EXPORT
-  char *get_llvm_cpu_name ();
+  char *pocl_get_llvm_cpu_name ();
   /* Returns if the cpu supports FMA instruction (uses LLVM). */
   int cpu_has_fma ();
 
@@ -68,6 +68,7 @@ extern "C" {
    * WG function that might be specialized according to the properties of
    * the given Command.
    */
+  POCL_EXPORT
   int pocl_llvm_generate_workgroup_function (unsigned DeviceI,
                                              cl_device_id Device,
                                              cl_kernel Kernel,
