@@ -67,6 +67,7 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 
 #include <iostream>
 #include <sstream>
+#include <regex>
 
 // For some reason including pocl.h before including CodeGenAction.h
 // causes an error. Some kind of macro definition issue. To investigate.
@@ -79,6 +80,7 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 #include "pocl_file_util.h"
 #include "pocl_cache.h"
 #include "LLVMUtils.h"
+#include "pocl_util.h"
 
 using namespace clang;
 using namespace llvm;
@@ -257,8 +259,6 @@ int pocl_llvm_build_program(cl_program program,
       POCL_MEM_FREE(device_switches);
     }
 
-  llvm::StringRef extensions(device->extensions);
-
 #if !(defined(__x86_64__) && defined(__GNUC__))
   if (program->flush_denorms) {
     POCL_MSG_WARN("flush to zero is currently only implemented for "
@@ -266,24 +266,6 @@ int pocl_llvm_build_program(cl_program program,
   }
 #endif
 
-  std::string cl_ext;
-  if (extensions.size() > 0) {
-    size_t e_start = 0, e_end = 0;
-    while (e_end < std::string::npos) {
-      e_end = extensions.find(' ', e_start);
-      llvm::StringRef tok = extensions.slice(e_start, e_end);
-      e_start = e_end + 1;
-      ss << "-D" << tok.str() << " ";
-      cl_ext += "+";
-      cl_ext += tok.str();
-      cl_ext += ",";
-    }
-  }
-  if (!cl_ext.empty()) {
-    cl_ext.back() = ' '; // replace last "," with space
-    ss << "-cl-ext=-all," << cl_ext;
-  }
-
   /* temp dir takes preference */
   if (num_input_headers > 0)
     ss << "-I" << temp_include_dir << " ";
@@ -293,9 +275,6 @@ int pocl_llvm_build_program(cl_program program,
 
   ss << "-DPOCL_DEVICE_ADDRESS_BITS=" << device->address_bits << " ";
   ss << "-D__USE_CLANG_OPENCL_C_H ";
-#ifndef LLVM_OLDER_THAN_13_0
-  ss << "-Dreserve_id_t=unsigned ";
-#endif
 
   ss << "-xcl ";
   // Remove the inline keywords to force the user functions
@@ -307,6 +286,28 @@ int pocl_llvm_build_program(cl_program program,
   // required for clGetKernelArgInfo()
   ss << "-cl-kernel-arg-info ";
 
+  size_t fastmath_flag = user_options.find("-cl-fast-relaxed-math");
+
+  if (fastmath_flag != std::string::npos) {
+#ifdef ENABLE_CONFORMANCE
+    user_options.replace(fastmath_flag, 21,
+                         "-cl-finite-math-only -cl-unsafe-math-optimizations");
+#endif
+    ss << "-D__FAST_RELAXED_MATH__=1 ";
+  }
+
+#ifdef ENABLE_CONFORMANCE
+  size_t unsafemath_flag = user_options.find("-cl-unsafe-math-optimizations");
+
+  if (unsafemath_flag != std::string::npos) {
+    // this should be almost the same but disables -freciprocal-math.
+    // required for conformance_math_divide test to pass with OpenCL 3.0
+    user_options.replace(unsafemath_flag, 29,
+                         "-cl-no-signed-zeros -cl-mad-enable -ffp-contract=fast");
+  }
+#endif
+
+
   ss << user_options << " ";
 
   if (device->endian_little)
@@ -317,13 +318,11 @@ int pocl_llvm_build_program(cl_program program,
 
   ss << "-DCL_DEVICE_MAX_GLOBAL_VARIABLE_SIZE=" << device->global_var_max_size << " ";
 
-  if (user_options.find("cl-fast-relaxed-math") != std::string::npos)
-    ss << "-D__FAST_RELAXED_MATH__=1 ";
 
-  ss << "-D__OPENCL_VERSION__=" << device->cl_version_int << " ";
+  ss << "-D__OPENCL_VERSION__=" << device->version_as_int << " ";
 
   if (user_options.find("-cl-std=") == std::string::npos)
-    ss << "-cl-std=" << device->cl_version_std << " ";
+    ss << "-cl-std=" << device->opencl_c_version_as_opt << " ";
 
   std::string temp(ss.str());
   size_t pos = temp.find("-cl-std=CL");
@@ -333,6 +332,35 @@ int pocl_llvm_build_program(cl_program program,
   int cl_std_i = cl_std_major * 100 + cl_std_minor * 10;
   ss << "-D__OPENCL_C_VERSION__=" << cl_std_i << " ";
 
+  std::string exts = device->extensions;
+  if (cl_std_major >= 3) {
+    exts += ' ';
+    exts += device->features;
+  }
+  llvm::StringRef extensions(exts);
+
+  std::string cl_ext;
+  if (extensions.size() > 0) {
+    size_t e_start = 0, e_end = 0;
+    while (e_end < std::string::npos) {
+      while (e_start < extensions.size() && std::isspace(extensions[e_start]))
+        ++e_start;
+      if (e_start >= extensions.size())
+        break;
+      e_end = extensions.find(' ', e_start);
+      llvm::StringRef tok = extensions.slice(e_start, e_end);
+      e_start = e_end + 1;
+      ss << "-D" << tok.str() << "=1 ";
+      cl_ext += "+";
+      cl_ext += tok.str();
+      cl_ext += ",";
+    }
+  }
+  if (!cl_ext.empty()) {
+    cl_ext.back() = ' '; // replace last "," with space
+    ss << "-cl-ext=-all," << cl_ext;
+  }
+
   ss << "-fno-builtin ";
   /* with fp-contract=on we get calls to fma with processors which do not
    * have fma instructions. These ruin the performance.
@@ -353,14 +381,45 @@ int pocl_llvm_build_program(cl_program program,
 
   POCL_MSG_PRINT_LLVM("all build options: %s\n", ss.str().c_str());
 
-  std::istream_iterator<std::string> begin(ss);
-  std::istream_iterator<std::string> end;
-  std::istream_iterator<std::string> i = begin;
-  std::vector<const char*> itemcstrs;
+  char WSReplacementChar = 0;
+
+  char *TempOptions = (char *) malloc (ss.str().length() + 1);
+
+  memset (TempOptions, 0, ss.str().length() + 1);
+  strncpy (TempOptions, ss.str().c_str(), ss.str().length());
+
+  if (pocl_escape_quoted_whitespace (TempOptions, &WSReplacementChar) == -1)
+  {
+    POCL_MEM_FREE (TempOptions);
+    return CL_INVALID_BUILD_OPTIONS;
+  }
+
+  std::istringstream iss(TempOptions);
+  std::vector<const char *> itemcstrs;
   std::vector<std::string> itemstrs;
-  while (i != end) {
-    itemstrs.push_back(*i);
-    ++i;
+
+  std::string s;
+
+  while (iss >> s)
+  {
+    // if needed, put back whitespace
+    if (WSReplacementChar != 0)
+    {
+      if (s.find(WSReplacementChar) != std::string::npos)
+      {
+        std::replace(s.begin(), s.end(), WSReplacementChar, ' ');
+      }
+    }
+
+    // if quoted, remove it to make compiler happy
+    if (s.find("\"") != std::string::npos)
+    {
+      std::regex Target("\"");
+      std::string Replacement = " ";
+      s = std::regex_replace(s, Target, Replacement);
+    }
+
+    itemstrs.push_back(s);
   }
 
   for (unsigned idx = 0; idx < itemstrs.size(); idx++) {
@@ -369,6 +428,8 @@ int pocl_llvm_build_program(cl_program program,
     itemcstrs.push_back(itemstrs[idx].c_str());
   }
 
+  POCL_MEM_FREE (TempOptions);
+
 #ifdef DEBUG_POCL_LLVM_API
   // TODO: for some reason the user_options are replicated,
   // they appear twice in a row in the output
@@ -652,7 +713,7 @@ int pocl_llvm_build_program(cl_program program,
    * save program->binaries[i]
    */
   std::string content;
-  writeModuleIR(mod, content);
+  writeModuleIRtoString(mod, content);
 
   if (program->binaries[device_i])
     POCL_MEM_FREE(program->binaries[device_i]);
@@ -802,11 +863,9 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
   if (error)
     return error;
 
-  /* To avoid writing & reading the same back,
-   * save program->binaries[i]
-   */
+  /* To avoid writing & reading the same back, save program->binaries[i] */
   std::string content;
-  writeModuleIR(linked_module, content);
+  writeModuleIRtoString(linked_module, content);
 
   if (program->binaries[device_i])
     POCL_MEM_FREE(program->binaries[device_i]);
@@ -876,7 +935,7 @@ static llvm::Module *getKernelLibrary(cl_device_id device,
   bool is_host = true;
   // TODO: move this to the device layer, a property to ask for
   // the kernel builtin bitcode library name, including its subdir
-#ifdef TCE_AVAILABLE
+#if defined(TCE_AVAILABLE)
   if (triple.getArch() == Triple::tce || triple.getArch() == Triple::tcele) {
     subdir = "tce";
     is_host = false;
@@ -922,7 +981,8 @@ static llvm::Module *getKernelLibrary(cl_device_id device,
     kernellib_fallback = kernellib;
     kernellib_fallback += OCL_KERNEL_TARGET_CPU;
     kernellib_fallback += ".bc";
-    kernellib += device->llvm_cpu;
+    if (device->llvm_cpu)
+      kernellib += device->llvm_cpu;
 #endif
   }
   kernellib += ".bc";
