@@ -75,9 +75,8 @@ llvm::Module *parseModuleIR(const char *path, llvm::LLVMContext *c) {
   return parseIRFile(path, Err, *c).release();
 }
 
-
-void writeModuleIR(const Module *mod, std::string &str) {
-  llvm::raw_string_ostream sos(str);
+void writeModuleIRtoString(const llvm::Module *mod, std::string& dest) {
+  llvm::raw_string_ostream sos(dest);
 #ifdef LLVM_OLDER_THAN_7_0
   WriteBitcodeToFile(mod, sos);
 #else
@@ -91,7 +90,7 @@ int pocl_write_module(void *module, const char *path, int dont_rewrite) {
   assert(path);
 
   std::string binary;
-  writeModuleIR((const Module *)module, binary);
+  writeModuleIRtoString((const Module *)module, binary);
 
   return pocl_write_file(path, binary.data(), (uint64_t)binary.size(), 0,
                          dont_rewrite);
@@ -122,8 +121,7 @@ static int getModuleTriple(const char *input_stream, size_t size,
   return 0;
 }
 
-char *
-get_llvm_cpu_name () {
+char *pocl_get_llvm_cpu_name() {
   StringRef r = llvm::sys::getHostCPUName();
 
   // LLVM may return an empty string -- treat as generic
@@ -345,6 +343,11 @@ void InitializeLLVM() {
       assert(O && "could not find LLVM option 'debug'");
       O->addOccurrence(1, StringRef("debug"), StringRef("true"), false);
     }
+#if LLVM_MAJOR == 9
+    O = opts["unroll-threshold"];
+    assert(O && "could not find LLVM option 'unroll-threshold'");
+    O->addOccurrence(1, StringRef("unroll-threshold"), StringRef("1"), false);
+#endif
   }
 }
 
