@@ -39,7 +39,11 @@
 IGNORE_COMPILER_WARNING("-Wunused-parameter")
 
 #include <llvm/Support/Casting.h>
+#ifdef LLVM_OLDER_THAN_14_0
 #include <llvm/Support/TargetRegistry.h>
+#else
+#include <llvm/MC/TargetRegistry.h>
+#endif
 #include <llvm/Support/SourceMgr.h>
 #include <llvm/Support/CommandLine.h>
 
@@ -198,7 +202,7 @@ static PassManager &kernel_compiler_passes(cl_device_id device) {
      Notes about the kernel compiler phase ordering:
 
      -mem2reg first because we get unoptimized output from Clang where all
-     variables are allocas. Avoid context saving the allocas and make the
+     variables are allocas. Avoid context saving the allocas and make them
      more readable by calling -mem2reg at the beginning.
 
      -implicit-cond-barriers after -implicit-loop-barriers because the latter
@@ -441,6 +445,8 @@ int pocl_llvm_generate_workgroup_function_nowrite(
   setModuleIntMetadata(ParallelBC, "device_address_bits", Device->address_bits);
   setModuleBoolMetadata(ParallelBC, "device_arg_buffer_launcher",
                         Device->arg_buffer_launcher);
+  setModuleBoolMetadata(ParallelBC, "device_grid_launcher",
+                        Device->grid_launcher);
   setModuleBoolMetadata(ParallelBC, "device_is_spmd", Device->spmd);
 
   setModuleStringMetadata(ParallelBC, "KernelName", Kernel->name);
@@ -603,7 +609,7 @@ int pocl_llvm_update_binaries(cl_program program, cl_uint device_i) {
   }
 
   std::string content;
-  writeModuleIR((llvm::Module *)program->data[i], content);
+  writeModuleIRtoString((llvm::Module *)program->data[i], content);
 
   size_t n = content.size();
   if (n < program->binary_sizes[i])
