@@ -42,14 +42,16 @@
 #  include "vccompat.hpp"
 #endif
 
-#include "pocl_util.h"
-#include "pocl_timing.h"
-#include "pocl_llvm.h"
-#include "utlist.h"
 #include "common.h"
-#include "pocl_mem_management.h"
 #include "devices.h"
+#include "pocl_cache.h"
+#include "pocl_file_util.h"
+#include "pocl_llvm.h"
+#include "pocl_mem_management.h"
 #include "pocl_runtime_config.h"
+#include "pocl_timing.h"
+#include "pocl_util.h"
+#include "utlist.h"
 
 /* required for setting SSE/AVX flush denorms to zero flag */
 #if defined(__x86_64__) && defined(__GNUC__)
@@ -167,7 +169,7 @@ pocl_restore_rm (unsigned rm)
 }
 
 uint32_t
-byteswap_uint32_t (uint32_t word, char should_swap)
+pocl_byteswap_uint32_t (uint32_t word, char should_swap)
 {
     union word_union
     {
@@ -202,6 +204,21 @@ byteswap_float (float word, char should_swap)
     return neww.full_word;
 }
 
+void
+bzero_s (void *v, size_t n)
+{
+  if (v == NULL)
+    return;
+
+  volatile unsigned char *p = v;
+  while (n--)
+    {
+      *p++ = 0;
+    }
+
+  return;
+}
+
 size_t
 pocl_size_ceil2(size_t x) {
   /* Rounds up to the next highest power of two without branching and
@@ -239,6 +256,17 @@ pocl_size_ceil2_64 (uint64_t x)
   return ++x;
 }
 
+/* Rounds up to the (power of two) alignment */
+size_t pocl_align_value (size_t value, size_t alignment)
+{
+  if (value & (alignment-1))
+    {
+      value |= (alignment-1);
+      ++value;
+    }
+  return value;
+}
+
 #if defined(_WIN32) || defined(HAVE_POSIX_MEMALIGN) || defined(__ANDROID__)    \
     || (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201112L))
 #define HAVE_ALIGNED_ALLOC
@@ -529,6 +557,15 @@ pocl_create_event_sync (cl_event waiting_event, cl_event notifier_event)
   LL_PREPEND (notifier_event->notify_list, notify_target);
   LL_PREPEND (waiting_event->wait_list, wait_list_item);
 
+  if (pocl_is_tracing_enabled ())
+    {
+      if (waiting_event->meta_data == NULL)
+        waiting_event->meta_data = (pocl_event_md *) calloc (1, sizeof (pocl_event_md));
+      pocl_event_md *md = waiting_event->meta_data;
+      if (md->num_deps < MAX_EVENT_DEPS)
+        md->dep_ids[md->num_deps++] = notifier_event->id;
+    }
+
 FINISH:
   pocl_unlock_events_inorder (waiting_event, notifier_event);
   return CL_SUCCESS;
@@ -1137,30 +1174,39 @@ pocl_command_push (_cl_command_node *node,
     }
 }
 
-void
-pocl_unmap_command_finished (cl_device_id dev, pocl_mem_identifier *mem_id,
-                             cl_mem mem, mem_mapping_t *map)
-{
-  POCL_LOCK_OBJ (mem);
-  assert (map->unmap_requested > 0);
-  dev->ops->free_mapping_ptr (dev->data, mem_id, mem, map);
-  DL_DELETE (mem->mappings, map);
-  mem->map_count--;
-  POCL_MEM_FREE (map);
-  POCL_UNLOCK_OBJ (mem);
-}
-
-void
-pocl_unmap_command_finished2 (cl_event event, _cl_command_t *cmd)
+static void
+pocl_unmap_command_finished (cl_event event, _cl_command_t *cmd)
 {
   cl_device_id dev = event->queue->device;
   pocl_mem_identifier *mem_id = NULL;
   cl_mem mem = NULL;
   mem = event->mem_objs[0];
   mem_id = &mem->device_ptrs[dev->global_mem_id];
-  pocl_unmap_command_finished (dev, mem_id, mem, cmd->unmap.mapping);
+
+  mem_mapping_t *map = cmd->unmap.mapping;
+  POCL_LOCK_OBJ (mem);
+  assert (map->unmap_requested > 0);
+  if (dev->ops->free_mapping_ptr)
+    dev->ops->free_mapping_ptr (dev->data, mem_id, mem, map);
+  DL_DELETE (mem->mappings, map);
+  mem->map_count--;
+  POCL_MEM_FREE (map);
+  POCL_UNLOCK_OBJ (mem);
 }
 
+static void
+pocl_ndrange_node_cleanup (_cl_command_node *node)
+{
+  cl_uint i;
+  for (i = 0; i < node->command.run.kernel->meta->num_args; ++i)
+    {
+      pocl_aligned_free (node->command.run.arguments[i].value);
+    }
+  POCL_MEM_FREE (node->command.run.arguments);
+  POname(clReleaseKernel)(node->command.run.kernel);
+}
+
+
 void
 pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer, cl_mem_flags flags)
 {
@@ -1527,7 +1573,7 @@ image_format_union (const cl_image_format *dev_formats,
 }
 
 /* Setup certain info about context that comes up later in API calls */
-void
+int
 pocl_setup_context (cl_context context)
 {
   unsigned i, j;
@@ -1571,6 +1617,9 @@ pocl_setup_context (cl_context context)
                 &context->image_formats[j], &context->num_image_formats[j]);
         }
 
+      if (dev->ops->init_context)
+        dev->ops->init_context (dev, context);
+
       context->default_queues[i] = POname (clCreateCommandQueue) (
           context, dev,
           (CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_HIDDEN
@@ -1582,6 +1631,23 @@ pocl_setup_context (cl_context context)
 
   assert (alignment > 0);
   context->min_buffer_alignment = alignment;
+  return CL_SUCCESS;
+}
+
+pocl_svm_ptr *
+pocl_find_svm_ptr_in_context (cl_context context, const void *host_ptr)
+{
+  POCL_LOCK_OBJ (context);
+  pocl_svm_ptr *item = NULL;
+  DL_FOREACH (context->svm_ptrs, item)
+  {
+    if (item->svm_ptr == host_ptr)
+      {
+        break;
+      }
+  }
+  POCL_UNLOCK_OBJ (context);
+  return item;
 }
 
 int
@@ -1629,8 +1695,8 @@ pocl_abort_on_pthread_error (int status, unsigned line, const char *func)
 {
   if (status != 0)
     {
-      POCL_MSG_PRINT2 (HSA, func, line, "Error from pthread call:\n");
-      POCL_ABORT ("%s\n", strerror (status));
+      POCL_MSG_PRINT2 (ERROR, func, line, "Error from pthread call:\n");
+      POCL_ABORT ("PTHREAD ERROR in %s():%u: %s (%d)\n", func, line, strerror (status), status);
     }
 }
 
@@ -1770,6 +1836,77 @@ pocl_run_command (char *const *args)
     }
 }
 
+int
+pocl_run_command_capture_output (char *capture_string, size_t *captured_bytes,
+                                 char *const *args)
+{
+  POCL_MSG_PRINT_INFO ("Launching: %s\n", args[0]);
+
+  int in[2];
+  int out[2];
+  pipe (in);
+  pipe (out);
+
+#ifdef HAVE_VFORK
+  pid_t p = vfork ();
+#elif defined(HAVE_FORK)
+  pid_t p = fork ();
+#else
+#error Must have fork() or vfork() system calls
+#endif
+  if (p == 0)
+    {
+      close (in[1]);
+      close (out[0]);
+
+      dup2 (in[0], STDIN_FILENO);
+      dup2 (out[1], STDOUT_FILENO);
+      dup2 (out[1], STDERR_FILENO);
+
+      return execv (args[0], args);
+    }
+  else
+    {
+      if (p < 0)
+        return EXIT_FAILURE;
+
+      close (in[0]);
+      close (out[1]);
+
+      ssize_t r = 0;
+      size_t total_bytes = 0;
+      size_t capture_limit = *captured_bytes;
+      char buf[4096];
+
+      while ((r = read (out[0], buf, 4096)) > 0)
+        {
+          if (total_bytes + r > capture_limit)
+            break;
+          memcpy (capture_string + total_bytes, buf, r);
+          total_bytes += r;
+        }
+      if (total_bytes > capture_limit)
+        total_bytes = capture_limit;
+
+      capture_string[total_bytes] = 0;
+      *captured_bytes = total_bytes;
+
+      int status;
+      if (waitpid (p, &status, 0) < 0)
+        POCL_ABORT ("pocl: waitpid() failed.\n");
+
+      close (out[0]);
+      close (in[1]);
+
+      if (WIFEXITED (status))
+        return WEXITSTATUS (status);
+      else if (WIFSIGNALED (status))
+        return WTERMSIG (status);
+      else
+        return EXIT_FAILURE;
+    }
+}
+
 // event locked
 void
 pocl_update_event_queued (cl_event event)
@@ -1836,10 +1973,62 @@ pocl_update_event_running (cl_event event)
   POCL_UNLOCK_OBJ (event);
 }
 
+static void pocl_free_event_node (cl_event event)
+{
+  _cl_command_node *node = event->command;
+  switch (node->type)
+    {
+    case CL_COMMAND_NDRANGE_KERNEL:
+    case CL_COMMAND_TASK:
+      pocl_ndrange_node_cleanup (node);
+      break;
+
+    case CL_COMMAND_FILL_BUFFER:
+      pocl_aligned_free (node->command.memfill.pattern);
+      break;
+
+    case CL_COMMAND_SVM_MEMFILL:
+      pocl_aligned_free (node->command.svm_fill.pattern);
+      break;
+
+    case CL_COMMAND_NATIVE_KERNEL:
+      POCL_MEM_FREE (node->command.native.args);
+      break;
+
+    case CL_COMMAND_UNMAP_MEM_OBJECT:
+      pocl_unmap_command_finished (event, &node->command);
+      break;
+
+    case CL_COMMAND_SVM_MIGRATE_MEM:
+      POCL_MEM_FREE (node->command.svm_migrate.sizes);
+      POCL_MEM_FREE (node->command.svm_migrate.svm_pointers);
+      break;
+    }
+  pocl_mem_manager_free_command (node);
+  event->command = NULL;
+}
+
+static void pocl_free_event_memobjs (cl_event event)
+{
+  size_t i;
+  for (i = 0; i < event->num_buffers; ++i)
+    {
+      cl_mem mem = event->mem_objs[i];
+      if (event->release_mem_host_ptr_after)
+        {
+          POCL_LOCK_OBJ (mem);
+          pocl_release_mem_host_ptr (mem);
+          POCL_UNLOCK_OBJ (mem);
+        }
+      POname (clReleaseMemObject) (mem);
+    }
+  POCL_MEM_FREE (event->mem_objs);
+}
+
 // status can be complete or failed (<0)
 void
-pocl_update_event_finished_msg (cl_int status, const char *func, unsigned line,
-                                cl_event event, const char *msg)
+pocl_update_event_finished (cl_int status, const char *func, unsigned line,
+                            cl_event event, const char *msg)
 {
   assert (event != NULL);
   assert (event->queue != NULL);
@@ -1872,12 +2061,6 @@ pocl_update_event_finished_msg (cl_int status, const char *func, unsigned line,
     cq->last_event.event = NULL;
   DL_DELETE (cq->events, event);
 
-  if (ops->notify_cmdq_finished && (cq->command_count == 0))
-    ops->notify_cmdq_finished (cq);
-
-  if (ops->notify_event_finished)
-    ops->notify_event_finished (event);
-
   POCL_UNLOCK_OBJ (cq);
   /* note that we must unlock the CmqQ before calling pocl_event_updated,
    * because it calls event callbacks, which can have calls to
@@ -1894,36 +2077,35 @@ pocl_update_event_finished_msg (cl_int status, const char *func, unsigned line,
     }
 #endif
 
-  size_t i;
-  for (i = 0; i < event->num_buffers; ++i)
-    {
-      cl_mem mem = event->mem_objs[i];
-      if (event->release_mem_host_ptr_after)
-        {
-          POCL_LOCK_OBJ (mem);
-          pocl_release_mem_host_ptr (mem);
-          POCL_UNLOCK_OBJ (mem);
-        }
-      POname (clReleaseMemObject) (mem);
-    }
-  POCL_MEM_FREE (event->mem_objs);
+  pocl_free_event_node (event);
+  pocl_free_event_memobjs (event);
+
+  POCL_LOCK_OBJ (cq);
+  if (ops->notify_cmdq_finished && (cq->command_count == 0))
+    ops->notify_cmdq_finished (cq);
+  POCL_UNLOCK_OBJ (cq);
+  POCL_LOCK_OBJ (event);
+  if (ops->notify_event_finished)
+    ops->notify_event_finished (event);
+  POCL_UNLOCK_OBJ (event);
 
   POname (clReleaseEvent) (event);
 }
 
+
 void
 pocl_update_event_failed (cl_event event)
 {
   POCL_UNLOCK_OBJ (event);
-  pocl_update_event_finished_msg (CL_FAILED, NULL, 0, event, NULL);
+  pocl_update_event_finished (CL_FAILED, NULL, 0, event, NULL);
   POCL_LOCK_OBJ (event);
 }
 
 void
-pocl_update_event_complete_msg (const char *func, unsigned line,
-                                cl_event event, const char *msg)
+pocl_update_event_complete (const char *func, unsigned line,
+                            cl_event event, const char *msg)
 {
-  pocl_update_event_finished_msg (CL_COMPLETE, func, line, event, msg);
+  pocl_update_event_finished (CL_COMPLETE, func, line, event, msg);
 }
 
 /*
@@ -2014,9 +2196,11 @@ float_to_half (float value)
 #define OpCapab 0x00020011
 /* execution model = Kernel is used by OpenCL SPIR-V modules */
 #define KernelExecModel 0x6
+/* execution model = Shader is used by Vulkan SPIR-V modules */
+#define ShaderExecModel 0x1
 
-int
-bitcode_is_spirv_kernel (const char *bitcode, size_t size)
+static int
+bitcode_is_spirv_execmodel (const char *bitcode, size_t size, uint32_t type)
 {
   const uint32_t *bc32 = (const uint32_t *)bitcode;
   unsigned location = 0;
@@ -2027,23 +2211,116 @@ bitcode_is_spirv_kernel (const char *bitcode, size_t size)
 
   // skip version, generator, bound, schema
   location += 4;
-  int is_opencl = 0;
-  uint32_t instruction, value;
-  do
+  int is_type = 0;
+  uint32_t value, instruction;
+  instruction = htole32 (bc32[location++]);
+  value = htole32 (bc32[location++]);
+  while (instruction == OpCapab && location < (size / 4))
     {
+      if (value == type)
+        return 1;
       instruction = htole32 (bc32[location++]);
       value = htole32 (bc32[location++]);
-      if (value == KernelExecModel)
-        is_opencl = 1;
-    }
-  while (instruction == OpCapab);
-
-  /* SPIR-V but not OpenCL-type. */
-  if (!is_opencl)
-    {
-      POCL_MSG_ERR ("SPIR-V binary provided, but is not using Kernel mode."
-                    "Pocl can't process this binary.\n");
     }
 
-  return is_opencl;
+  return 0;
+}
+
+int
+pocl_bitcode_is_spirv_execmodel_kernel (const char *bitcode, size_t size)
+{
+  return bitcode_is_spirv_execmodel (bitcode, size, KernelExecModel);
+}
+
+int
+pocl_bitcode_is_spirv_execmodel_shader (const char *bitcode, size_t size)
+{
+  return bitcode_is_spirv_execmodel (bitcode, size, ShaderExecModel);
+}
+
+int
+pocl_device_is_associated_with_kernel (cl_device_id device, cl_kernel kernel)
+{
+  unsigned i;
+  int found_it = 0;
+  for (i = 0; i < kernel->context->num_devices; i++)
+    if (pocl_real_dev (device) == kernel->context->devices[i])
+      {
+        found_it = 1;
+        break;
+      }
+
+  return found_it;
+}
+
+/*
+ * search for an unused ASCII character in options,
+ * to be used to replace whitespaces within double quoted substrings
+ */
+static int
+pocl_find_unused_char (const char *options, char *replace_me)
+{
+  for (int y = 35; y < 128; y++)
+  {
+    if (strchr (options, (char) y) == NULL)
+    {
+      *replace_me = (char) y;
+      return 0;
+    }
+  }
+
+  return -1;
+}
+
+int
+pocl_escape_quoted_whitespace (char *temp_options, char *replace_me)
+{
+  /* searching for double quote in temp_options */
+  if (strchr (temp_options, '"') != NULL)
+  {
+    size_t replace_cnt = 0;
+
+    int in_substring = -1;
+
+    /* scan for double quoted substring */
+    for (size_t x = 0; x < strlen (temp_options); x++)
+    {
+      if (temp_options[x] == '"')
+      {
+        if (in_substring == -1)
+        {
+          /* enter in double quoted substring */
+          in_substring = 0;
+          continue;
+        }
+
+        /* exit from double quoted substring */
+        in_substring = -1;
+        continue;
+      }
+
+      /* search for whitespaces in substring */
+      if (in_substring == 0)
+      {
+        if (temp_options[x] == ' ')
+        {
+          /* at first need, get an unused char */
+          if (replace_cnt == 0)
+          {
+            if (pocl_find_unused_char (temp_options, replace_me) == -1)
+            {
+              /* no replace, no party */
+              return -1;
+            }
+          }
+
+          /* replace whitespace with unused char */
+          temp_options[x] = *replace_me;
+          replace_cnt++;
+        }
+      }
+    }
+  }
+
+  return 0;
 }
