@@ -33,7 +33,8 @@
 extern "C" {
 #endif
 
-uint32_t byteswap_uint32_t (uint32_t word, char should_swap);
+POCL_EXPORT
+uint32_t pocl_byteswap_uint32_t (uint32_t word, char should_swap);
 float byteswap_float (float word, char should_swap);
 
 /* set rounding mode */
@@ -59,9 +60,14 @@ void pocl_restore_ftz (unsigned ftz);
 void pocl_install_sigfpe_handler ();
 void pocl_install_sigusr2_handler ();
 
+void bzero_s (void *v, size_t n);
+
 /* Finds the next highest power of two of the given value. */
+POCL_EXPORT
 size_t pocl_size_ceil2 (size_t x);
 uint64_t pocl_size_ceil2_64 (uint64_t x);
+POCL_EXPORT
+size_t pocl_align_value (size_t value, size_t alignment);
 
 /* Allocates aligned blocks of memory.
  *
@@ -107,8 +113,10 @@ cl_int pocl_create_command_migrate (_cl_command_node **cmd,
 void pocl_command_enqueue (cl_command_queue command_queue,
                           _cl_command_node *node);
 
+POCL_EXPORT
 int pocl_alloc_or_retain_mem_host_ptr (cl_mem mem);
 
+POCL_EXPORT
 int pocl_release_mem_host_ptr (cl_mem mem);
 
 /* does several sanity checks on buffer & given memory region */
@@ -124,6 +132,9 @@ int pocl_buffer_boundcheck_3d(const size_t buffer_size, const size_t *origin,
                               const size_t *region, size_t *row_pitch,
                               size_t *slice_pitch, const char* prefix);
 
+pocl_svm_ptr *pocl_find_svm_ptr_in_context (cl_context context,
+                                            const void *host_ptr);
+
 int
 check_copy_overlap(const size_t src_offset[3],
                    const size_t dst_offset[3],
@@ -140,12 +151,6 @@ pocl_command_push (_cl_command_node *node,
                    _cl_command_node **ready_list,
                    _cl_command_node **pending_list);
 
-void pocl_unmap_command_finished (cl_device_id dev,
-                                  pocl_mem_identifier *mem_id, cl_mem mem,
-                                  mem_mapping_t *map);
-
-void pocl_unmap_command_finished2 (cl_event event, _cl_command_t *cmd);
-
 /**
  * Return true if a command is ready to execute (no more event in wait list
  * or false if not
@@ -161,7 +166,7 @@ typedef void (*empty_queue_callback) (cl_command_queue cq);
 void pocl_cl_mem_inherit_flags (cl_mem mem, cl_mem from_buffer,
                                 cl_mem_flags flags);
 
-void pocl_setup_context(cl_context context);
+int pocl_setup_context (cl_context context);
 
 /* Helpers for dealing with devices / subdevices */
 
@@ -187,30 +192,26 @@ int pocl_check_event_wait_list(cl_command_queue     command_queue,
                                cl_uint              num_events_in_wait_list,
                                const cl_event *     event_wait_list);
 
-void pocl_abort_on_pthread_error (int status, unsigned line, const char *func);
-
-#define PTHREAD_CHECK(code)                                                   \
-  pocl_abort_on_pthread_error ((code), __LINE__, __FUNCTION__);
-
 void pocl_update_event_queued (cl_event event);
 
 POCL_EXPORT
 void pocl_update_event_submitted (cl_event event);
 
+POCL_EXPORT
 void pocl_update_event_running_unlocked (cl_event event);
 
 POCL_EXPORT
 void pocl_update_event_running (cl_event event);
 
 POCL_EXPORT
-void pocl_update_event_complete_msg (const char *func, unsigned line,
-                                     cl_event event, const char *msg);
+void pocl_update_event_complete (const char *func, unsigned line,
+                                 cl_event event, const char *msg);
 
 #define POCL_UPDATE_EVENT_COMPLETE_MSG(__event, msg)                          \
-  pocl_update_event_complete_msg (__func__, __LINE__, (__event), msg);
+  pocl_update_event_complete (__func__, __LINE__, (__event), msg);
 
 #define POCL_UPDATE_EVENT_COMPLETE(__event)                                   \
-  pocl_update_event_complete_msg (__func__, __LINE__, (__event), NULL);
+  pocl_update_event_complete (__func__, __LINE__, (__event), NULL);
 
 POCL_EXPORT
 void pocl_update_event_failed (cl_event event);
@@ -218,17 +219,38 @@ void pocl_update_event_failed (cl_event event);
 const char*
 pocl_status_to_str (int status);
 
+POCL_EXPORT
 const char *
 pocl_command_to_str (cl_command_type cmd);
 
+POCL_EXPORT
 int
 pocl_run_command(char * const *args);
 
+POCL_EXPORT
+int pocl_run_command_capture_output (char *capture_string,
+                                     size_t *captured_bytes,
+                                     char *const *args);
+
 uint16_t float_to_half (float value);
 
 float half_to_float (uint16_t value);
 
-int bitcode_is_spirv_kernel (const char *bitcode, size_t size);
+/* returns !0 if binary is SPIR-V bitcode with OpCapability Kernel
+ * OpenCL-style bitcode produced by e.g. llvm-spirv */
+POCL_EXPORT
+int pocl_bitcode_is_spirv_execmodel_kernel (const char *bitcode, size_t size);
+
+/* returns !0 if binary is SPIR-V bitcode with OpCapability Shader
+ * these are produced by e.g. GLSL and clspv compilation */
+POCL_EXPORT
+int pocl_bitcode_is_spirv_execmodel_shader (const char *bitcode, size_t size);
+
+int pocl_device_is_associated_with_kernel (cl_device_id device,
+                                           cl_kernel kernel);
+
+int pocl_escape_quoted_whitespace (char *temp_options, char *replace_me);
+
 
 #ifdef __cplusplus
 }
