@@ -55,8 +55,7 @@ using namespace pocl;
 
 int ParallelRegion::idGen = 0;
 
-ParallelRegion::ParallelRegion(int forcedRegionId) : 
-  std::vector<llvm::BasicBlock *>(), 
+ParallelRegion::ParallelRegion(int forcedRegionId) :
   LocalIDXLoadInstr(NULL), LocalIDYLoadInstr(NULL), LocalIDZLoadInstr(NULL),
   exitIndex_(0), entryIndex_(0), pRegionId(forcedRegionId)
 {
@@ -623,9 +622,12 @@ ParallelRegion::LocalIDZLoad()
 {
   if (LocalIDZLoadInstr != NULL) return LocalIDZLoadInstr;
   IRBuilder<> builder(&*(entryBB()->getFirstInsertionPt()));
-  return LocalIDZLoadInstr =
-    builder.CreateLoad
-    (entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_Z_GLOBAL));
+  Value *Ptr = entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_Z_GLOBAL);
+  return LocalIDZLoadInstr = builder.CreateLoad(
+#ifndef LLVM_OLDER_THAN_13_0
+    Ptr->getType()->getPointerElementType(),
+#endif
+    Ptr);
 }
 
 /**
@@ -637,9 +639,12 @@ ParallelRegion::LocalIDYLoad()
 {
   if (LocalIDYLoadInstr != NULL) return LocalIDYLoadInstr;
   IRBuilder<> builder(&*(entryBB()->getFirstInsertionPt()));
-  return LocalIDYLoadInstr = 
-    builder.CreateLoad
-    (entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_Y_GLOBAL));
+  Value *Ptr = entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_Y_GLOBAL);
+  return LocalIDYLoadInstr = builder.CreateLoad(
+#ifndef LLVM_OLDER_THAN_13_0
+    Ptr->getType()->getPointerElementType(),
+#endif
+    Ptr);
 }
 
 /**
@@ -651,9 +656,12 @@ ParallelRegion::LocalIDXLoad()
 {
   if (LocalIDXLoadInstr != NULL) return LocalIDXLoadInstr;
   IRBuilder<> builder(&*(entryBB()->getFirstInsertionPt()));
-  return LocalIDXLoadInstr = 
-    builder.CreateLoad
-    (entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_X_GLOBAL));
+  Value *Ptr = entryBB()->getParent()->getParent()->getGlobalVariable(POCL_LOCAL_ID_X_GLOBAL);
+  return LocalIDXLoadInstr = builder.CreateLoad(
+#ifndef LLVM_OLDER_THAN_13_0
+    Ptr->getType()->getPointerElementType(),
+#endif
+    Ptr);
 }
 
 void
@@ -688,10 +696,18 @@ ParallelRegion::InjectPrintF
        /*Name=*/"printf", M); 
     printfFunc->setCallingConv(CallingConv::C);
 
+#ifndef LLVM_OLDER_THAN_14_0
     AttributeList func_printf_PAL =
-      AttributeList()
-      .addAttribute(M->getContext(), 1U, Attribute::NoCapture)
-      .addAttribute(M->getContext(), 4294967295U, Attribute::NoUnwind);
+        AttributeList()
+            .addAttributeAtIndex(M->getContext(), 1U, Attribute::NoCapture)
+            .addAttributeAtIndex(M->getContext(), 4294967295U,
+                                 Attribute::NoUnwind);
+#else
+    AttributeList func_printf_PAL =
+        AttributeList()
+            .addAttribute(M->getContext(), 1U, Attribute::NoCapture)
+            .addAttribute(M->getContext(), 4294967295U, Attribute::NoUnwind);
+#endif
 
     printfFunc->setAttributes(func_printf_PAL);
   }
