@@ -2,7 +2,7 @@
 // and parallelized kernel for an OpenCL workgroup.
 //
 // Copyright (c) 2011 Universidad Rey Juan Carlos
-//               2012-2019 Pekka Jääskeläinen
+//               2012-2022 Pekka Jääskeläinen
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -170,6 +170,8 @@ Workgroup::runOnModule(Module &M) {
   getModuleIntMetadata(M, "device_address_bits", address_bits);
   getModuleBoolMetadata(M, "device_arg_buffer_launcher",
                         DeviceUsingArgBufferLauncher);
+  getModuleBoolMetadata(M, "device_grid_launcher",
+                        DeviceUsingGridLauncher);
   getModuleBoolMetadata(M, "device_is_spmd", DeviceIsSPMD);
 
   getModuleStringMetadata(M, "KernelName", KernelName);
@@ -267,7 +269,8 @@ Workgroup::runOnModule(Module &M) {
       L->addFnAttr(Attribute::NoInline);
       L->removeFnAttr(Attribute::AlwaysInline);
       WGLauncher->addFnAttr(Attribute::AlwaysInline);
-      createGridLauncher(L, WGLauncher, OrigKernel.getName().str());
+      if (DeviceUsingGridLauncher)
+        createGridLauncher(L, WGLauncher, OrigKernel.getName().str());
     } else if (DeviceIsSPMD) {
       // For SPMD machines there is no need for a WG launcher, the device will
       // call/handle the single-WI kernel function directly.
@@ -402,7 +405,7 @@ llvm::Value *
 Workgroup::createLoadFromContext(
   IRBuilder<> &Builder, int StructFieldIndex, int FieldIndex=-1) {
 
-  Value *GEP;
+  Value *GEP, *Ptr;
   Type *ContextType = ContextArg->getType()->getPointerElementType();
   GEP = Builder.CreateStructGEP(ContextType, ContextArg, StructFieldIndex);
   Type *GEPType = GEP->getType()->getPointerElementType();
@@ -410,29 +413,34 @@ Workgroup::createLoadFromContext(
   llvm::LoadInst *Load = nullptr;
   if (SizeTWidth == 64) {
     if (FieldIndex == -1)
-      Load = Builder.CreateLoad(Builder.CreateConstGEP1_64(
+      Ptr = Builder.CreateConstGEP1_64(
 #ifndef LLVM_OLDER_THAN_13_0
           GEPType,
 #endif
-          GEP, 0));
+          GEP, 0);
     else
-      Load = Builder.CreateLoad(Builder.CreateConstGEP2_64(
+      Ptr = Builder.CreateConstGEP2_64(
 #ifndef LLVM_OLDER_THAN_13_0
           GEPType,
 #endif
-          GEP, 0, FieldIndex));
+          GEP, 0, FieldIndex);
   } else {
     if (FieldIndex == -1)
-      Load = Builder.CreateLoad(Builder.CreateConstGEP1_32(
+      Ptr = Builder.CreateConstGEP1_32(
 #ifndef LLVM_OLDER_THAN_13_0
           GEPType,
 #endif
-          GEP, 0));
+          GEP, 0);
     else
-      Load = Builder.CreateLoad(Builder.CreateConstGEP2_32(
+      Ptr = Builder.CreateConstGEP2_32(
           GEPType,
-          GEP, 0, FieldIndex));
+          GEP, 0, FieldIndex);
   }
+  Load = Builder.CreateLoad(
+#ifndef LLVM_OLDER_THAN_13_0
+	  Ptr->getType()->getPointerElementType(),
+#endif
+	  Ptr);
   addRangeMetadataForPCField(Load, StructFieldIndex, FieldIndex);
   return Load;
 }
@@ -557,8 +565,22 @@ static void replacePrintfCalls(Value *pb, Value *pbp, Value *pbc, bool isKernel,
         ops.push_back(pbc);
 
         unsigned j = CallInstr->getNumOperands() - 1;
-        for (unsigned i = 0; i < j; ++i)
-          ops.push_back(CallInstr->getOperand(i));
+        for (unsigned i = 0; i < j; ++i) {
+          auto *Operand = CallInstr->getOperand(i);
+#ifndef LLVM_OLDER_THAN_10_0
+          // The printf decl might have the format string in the constant AS
+          // in order to support compilation from SPIR-V where the calls adhere
+          // to the SPIR-V/OpenCL standard in terms of the argument type.
+          // Thus, when compiling directly from OpenCL C to native LLVM we
+          // have to add an (temporarily illegal) AS cast in case the target
+          // is a flat address space target (CPUs).
+          if (i == 0)
+            Operand = llvm::CastInst::CreatePointerBitCastOrAddrSpaceCast(
+                Operand, poclPrintf->getArg(3)->getType(),
+                "printf_fmt_str_as_cast", CallInstr);
+#endif
+          ops.push_back(Operand);
+        }
 
         CallInst *NewCI = CallInst::Create(poclPrintf, ops);
         NewCI->setCallingConv(poclPrintf->getCallingConv());
@@ -1135,6 +1157,7 @@ static size_t getArgumentSize(llvm::Argument &Arg) {
   return DL.getTypeStoreSize(TypeInBuf);
 }
 
+// Tofix: Why this is duplicated here? pocl_utils.c should be used?
 static uint64_t pocl_size_ceil2_64(uint64_t x) {
   /* Rounds up to the next highest power of two without branching and
    * is as fast as a BSR instruction on x86, see:
@@ -1207,7 +1230,12 @@ Workgroup::createAllocaMemcpyForStruct(LLVMModuleRef M, LLVMBuilderRef Builder,
     args[1] = CARG1;
     args[2] = Size;
 
-    LLVMValueRef call4 = LLVMBuildCall(Builder, MemCpy4, args, 3, "");
+#ifdef LLVM_OLDER_THAN_14_0
+    LLVMValueRef Call4 = LLVMBuildCall(Builder, MemCpy4, args, 3, "");
+#else
+    LLVMTypeRef FnTy = LLVMGetCalledFunctionType(MemCpy4);
+    LLVMValueRef Call4 = LLVMBuildCall2(Builder, FnTy, MemCpy4, args, 3, "");
+#endif
   } else {
     LLVMTypeRef i8PtrAS0 = LLVMPointerType(Int8Type, 0);
     LLVMTypeRef i8PtrAS1 = LLVMPointerType(Int8Type, DeviceArgsASid);
@@ -1221,7 +1249,12 @@ Workgroup::createAllocaMemcpyForStruct(LLVMModuleRef M, LLVMBuilderRef Builder,
     args[1] = CARG1;
     args[2] = Size;
 
-    LLVMValueRef call1 = LLVMBuildCall(Builder, MemCpy1, args, 3, "");
+#ifdef LLVM_OLDER_THAN_14_0
+    LLVMValueRef Call1 = LLVMBuildCall(Builder, MemCpy1, args, 3, "");
+#else
+    LLVMTypeRef FnTy = LLVMGetCalledFunctionType(MemCpy1);
+    LLVMValueRef Call1 = LLVMBuildCall2(Builder, FnTy, MemCpy1, args, 3, "");
+#endif
   }
 
   return LocalArgAlloca;
@@ -1262,10 +1295,15 @@ LLVMValueRef Workgroup::createArgBufferLoad(LLVMBuilderRef Builder,
 
     // not by-val argument
   } else {
-    LLVMValueRef ArgOffsetBitcast = LLVMBuildPointerCast(
-        Builder, ArgByteOffset, LLVMPointerType(ParamType, DeviceArgsASid),
-        "arg_ptr");
+    LLVMTypeRef DestTy = LLVMPointerType(ParamType, DeviceArgsASid);
+    LLVMValueRef ArgOffsetBitcast =
+        LLVMBuildPointerCast(Builder, ArgByteOffset, DestTy, "arg_ptr");
+#ifdef LLVM_OLDER_THAN_14_0
     return LLVMBuildLoad(Builder, ArgOffsetBitcast, "");
+#else
+    LLVMTypeRef LoadTy = LLVMGetElementType(DestTy);
+    return LLVMBuildLoad2(Builder, LoadTy, ArgOffsetBitcast, "");
+#endif
   }
 }
 
@@ -1375,14 +1413,20 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
         LLVMValueRef Offs = LLVMConstInt(Int32Type, ArgPos, 0);
         LLVMValueRef SizeByteOffset =
             LLVMBuildGEP(Builder, ArgBuffer, &Offs, 1, "size_byte_offset");
+        LLVMTypeRef DestTy = LLVMPointerType(ParamIntType, 0);
         LLVMValueRef SizeOffsetBitcast =
-            LLVMBuildPointerCast(Builder, SizeByteOffset,
-                                 LLVMPointerType(ParamIntType, 0), "size_ptr");
+            LLVMBuildPointerCast(Builder, SizeByteOffset, DestTy, "size_ptr");
 
         // The buffer size passed from the runtime is a byte size, we
         // need to convert it to an element count for the alloca.
+#ifdef LLVM_OLDER_THAN_14_0
         LLVMValueRef LocalArgByteSize =
             LLVMBuildLoad(Builder, SizeOffsetBitcast, "byte_size");
+#else
+        LLVMTypeRef LoadTy = LLVMGetElementType(DestTy);
+        LLVMValueRef LocalArgByteSize =
+            LLVMBuildLoad2(Builder, LoadTy, SizeOffsetBitcast, "byte_size");
+#endif
         uint64_t ElementSize = LLVMStoreSizeOfType(DataLayout, ArgElementType);
         LLVMValueRef ElementCount =
             LLVMBuildUDiv(Builder, LocalArgByteSize,
@@ -1425,7 +1469,12 @@ Workgroup::createArgBufferWorkgroupLauncher(Function *Func,
 
   assert (i == ArgCount);
 
+#ifdef LLVM_OLDER_THAN_14_0
   LLVMValueRef Call = LLVMBuildCall(Builder, F, Args, ArgCount, "");
+#else
+  LLVMTypeRef FnTy = wrap(Func->getFunctionType());
+  LLVMValueRef Call = LLVMBuildCall2(Builder, FnTy, F, Args, ArgCount, "");
+#endif
   LLVMBuildRetVoid(Builder);
 
   llvm::CallInst *CallI = llvm::dyn_cast<llvm::CallInst>(llvm::unwrap(Call));
@@ -1511,7 +1560,12 @@ Workgroup::createGridLauncher(Function *KernFunc, Function *WGFunc,
       LLVMBuildPointerCast(Builder, PoclCtx, ArgTypes[2], "ctx"),
       LLVMBuildPointerCast(Builder, AuxParam, ArgTypes[1], "aux")};
 
+#ifdef LLVM_OLDER_THAN_14_0
   LLVMValueRef Call = LLVMBuildCall(Builder, RunnerFunc, Args, 4, "");
+#else
+  LLVMTypeRef FnTy = LLVMGetCalledFunctionType(RunnerFunc);
+  LLVMValueRef Call = LLVMBuildCall2(Builder, FnTy, RunnerFunc, Args, 4, "");
+#endif
   LLVMBuildRetVoid(Builder);
 
   InlineFunctionInfo IFI;
