@@ -21,8 +21,6 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#define DEBUG_TYPE "workitem-loops"
-
 #include <iostream>
 #include <map>
 #include <sstream>
@@ -48,6 +46,8 @@ IGNORE_COMPILER_WARNING("-Wunused-parameter")
 #include "llvm/Support/CommandLine.h"
 #include "llvm/Transforms/Utils/BasicBlockUtils.h"
 
+#define DEBUG_TYPE "workitem-loops"
+
 #include "WorkitemLoops.h"
 #include "Workgroup.h"
 #include "Barrier.h"
@@ -378,7 +378,7 @@ WorkitemLoops::ProcessFunction(Function &F)
   Initialize(K);
   unsigned workItemCount = WGLocalSizeX*WGLocalSizeY*WGLocalSizeZ;
 
-  if (workItemCount == 1)
+  if (workItemCount == 1 && !WGDynamicLocalSize)
     {
       K->addLocalSizeInitCode(WGLocalSizeX, WGLocalSizeY, WGLocalSizeZ);
       ParallelRegion::insertLocalIdInit(&F.getEntryBlock(), 0, 0, 0);
