@@ -1,7 +1,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <CL/opencl.h>
+
 #include "poclu.h"
 
 #ifdef _MSC_VER
@@ -139,5 +139,5 @@ int main(int argc, char **argv)
   free (imageData);
 
   printf("OK\n");
-  return 0;
+  return EXIT_SUCCESS;
 }
