@@ -1,7 +1,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <CL/opencl.h>
 
 #include "poclu.h"
 
@@ -76,6 +75,7 @@ int call_test(const char *name)
   result = clBuildProgram(program, 0, NULL, "-I" SRCDIR, NULL, NULL);
   if (result != CL_SUCCESS) {
     puts("clBuildProgram call failed\n");
+    poclu_show_program_build_log (program);
     goto error;
   }
 
@@ -167,6 +167,5 @@ int main(int argc, char **argv)
     printf("OK\n");
   fflush(stdout);
   fflush(stderr);
-  return (retval ? 1 : 0);
+  return (retval ? EXIT_FAILURE : EXIT_SUCCESS);
 }
-
