@@ -1,7 +1,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <CL/opencl.h>
+
 #include "poclu.h"
 
 #ifdef _MSC_VER
@@ -182,9 +182,9 @@ error:
   if (retval) 
     {
       printf("FAIL\n");
-      return 1;
+      return EXIT_FAILURE;
     }
  
   printf("OK\n");
-  return 0;
+  return EXIT_SUCCESS;
 }
