@@ -22,11 +22,12 @@
    THE SOFTWARE.
 */
 
+#include "pocl_opencl.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <CL/cl.h>
-#include <poclu.h>
+
 
 #define MAX_PLATFORMS 32
 #define MAX_DEVICES   32
@@ -67,5 +68,6 @@ main(void){
 
   CHECK_CL_ERROR (clUnloadCompiler ());
 
+  printf ("OK\n");
   return EXIT_SUCCESS;
 }
