@@ -5,9 +5,15 @@ if [ ! -e .git ]; then
   exit 1
 fi
 
-PATCHY=/tmp/p.patch
+case "$(git describe --always --dirty=-DIRTY)" in
+  *-DIRTY)
+    echo "There are uncommitted changes - aborting."
+    exit 1
+esac
+
+PATCHY=$(mktemp /tmp/pocl.XXXXXXXX.patch)
+trap "rm -f $PATCHY" EXIT
 
-rm -f $PATCHY
 git show -U0 --no-color >$PATCHY
 
 SCRIPTPATH=$( realpath "$0"  )
@@ -16,6 +22,11 @@ RELPATH=$(dirname "$SCRIPTPATH")
 $RELPATH/clang-format-diff.py -regex '.*(\.h$|\.c$|\.cl$)' -i -p1 -style GNU <$PATCHY
 $RELPATH/clang-format-diff.py -regex '(.*(\.hh$|\.cc$))|(lib/llvmopencl/.*\.h)' -i -p1 -style LLVM <$PATCHY
 
+if [ -z "$(git diff)" ]; then
+  echo "No changes."
+  exit 0
+fi
+
 git diff
 
 echo "ACCEPT CHANGES ?"
@@ -28,7 +39,7 @@ if [ "$REPLY" == "y" ]; then
 
   git commit --amend
 
-  if [ -e .git/ORIG_HEAD ]; then
+  if [ -d .git/rebase-merge ]; then
 
     git rebase --continue
 
